
import os
import subprocess
import shutil
from timeit import default_timer as timer

import logging
log = logging.getLogger(__name__)

from partis.schema import (
  required,
  optional,
  derived,
  is_sequence,
  is_mapping,
  is_evaluated,
  is_valued,
  is_valued_type,
  is_optional,
  PyEvaluated,
  CheetahEvaluated,
  BoolPrim,
  IntPrim,
  FloatPrim,
  StrPrim,
  SeqPrim,
  MapPrim,
  UnionPrim,
  StructValued,
  schema_declared,
  EvaluatedContext )

from partis.utils import (
  ModelHint,
  HINT_LEVELS_DESC,
  indent_lines )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
hint_declared = schema_declared( tag = 'hint' )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
HintList = SeqPrim(
  doc = "List of hints",
  item = hint_declared,
  default_val = list() )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class Hint( StructValued ):
  """A schema for serializing ModelHints
  """

  schema = dict(
    declared = hint_declared,
    default_val = optional,
    struct_proxy = 'msg' )

  level = StrPrim(
    char_case = 'upper',
    default_val = "INFO",
    restricted = [ k.upper() for k,v in HINT_LEVELS_DESC.items() ],
    doc = "\n".join([ f"- ``'{k.upper()}'``: {indent_lines(2, v, start = 1)}" for k,v in HINT_LEVELS_DESC.items() ]) )

  msg = StrPrim(
    doc = "Message",
    default_val = "" )

  loc = StrPrim(
    doc = "Referenced location",
    default_val = "",
    max_lines = 1 )

  hints = HintList

  #-----------------------------------------------------------------------------
  def _cast( self ):
    """Converts instance of this to instance of a regular ModelHint

    Returns
    :class:`ModelHint <partis.utils.hint.ModelHint>`
    """
    return ModelHint(
      self.msg,
      loc = self.loc,
      level = self.level,
      hints = [ h._cast() if h is not None else None for h in self.hints ] )



import pytest
import numpy as np
import array

import itertools
from collections.abc import (
  Mapping )

from partis.utils import (
  odict )

from partis.schema import (
  required,
  optional,
  derived,
  BoolPrim,
  IntPrim,
  FloatPrim,
  StrPrim,
  SeqPrim,
  MapPrim,
  UnionPrim,
  StructValued,
  PJCEvaluated,
  SchemaError,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_declared,
  is_schema_struct_valued,
  is_valued_type,
  is_valued,
  is_evaluated,
  is_similar_value_type,
  any_schema,
  schema_declared,
  SchemaDefinitionError,
  SchemaDeclaredError,
  PJCEvaluated,
  StructValued,
  SchemaStructProxy )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
namespace_combos = [
  ( 'tag_key', 'key1', 'key2' ),
  ( 'tag', 'tag1', 'tag2' ),
  ( 'struct_proxy', 'x1', 'x2' ),
  ( 'declared', schema_declared( tag = 'tag1' ), schema_declared( tag = 'tag2' ) ),
  ( 'evaluated', PJCEvaluated, None ),
  ( 'default_val', { 'x1': True, 'x2' : True }, { 'x1': True, 'x2' : False } ),
  ( 'default_eval', { 'x1': True, 'x2' : True }, { 'x1': True, 'x2' : False } ),
  ( 'init_val', { 'x1': True, 'x2' : True }, { 'x1': True, 'x2' : False } ) ]

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_namespace():

  # should pass
  for i in range(len(namespace_combos)):
    k, a, b = namespace_combos[i]

    d1 = {
      k : a }

    d2 = {
      k : a }

    if k == 'default_eval':
      # need to set as evaluated if setting default_eval
      d1['evaluated'] = PJCEvaluated
      d1['default_val'] = "$expr:py { 'x1' : True }"

      d2['evaluated'] = PJCEvaluated
      d2['default_val'] = "$expr:py { 'x1' : True }"

    class MyType1( StructValued, **d1 ):
      schema = d2
      x1 = BoolPrim()
      x2 = BoolPrim( default_val = True )


  class MyType1( StructValued ):
    schema = dict(
      struct_proxy = 'x1' )

    x1 = BoolPrim()
    x2 = BoolPrim( default_val = True )

  # should fail
  with pytest.raises( SchemaDefinitionError ):
    # must be mapping
    class MyType1( StructValued ):
      schema = list()
      pass

  # conflicting
  for i in range(len(namespace_combos)):
    k, a, b = namespace_combos[i]

    d1 = {
      k : a }

    d2 = {
      k : b }

    with pytest.raises( SchemaDefinitionError ):
      class MyType1( StructValued, **d1 ):
        schema = d2
        x1 = BoolPrim()
        x2 = BoolPrim( default_val = True )


  with pytest.raises( SchemaDefinitionError ):
    # proxy must be in schema
    class MyType1( StructValued ):
      schema = dict(
        struct_proxy = 'x1' )

  with pytest.raises( SchemaDefinitionError ):
    # proxy can be only non-default
    class MyType1( StructValued ):
      schema = dict(
        struct_proxy = 'x1' )
      x1 = BoolPrim()
      x2 = BoolPrim()

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_declared():


  d1 = schema_declared( tag = 'my_type1', tag_key = 'mytagkey' )

  class MyType1( StructValued,
    declared = d1,
    tag = 'my_type1',
    tag_key = 'mytagkey',
    struct = [ ] ):
    pass

  # already defined
  with pytest.raises( SchemaDeclaredError ):

    class MyType2( StructValued,
      declared = d1,
      struct = [ ] ):
      pass

  # type mismatch
  with pytest.raises( SchemaDefinitionError ):

    class MyType2( StructValued,
      declared = MyType1,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      struct = [ ] ):
      pass

  # tag mismatch
  with pytest.raises( SchemaDefinitionError ):
    d1 = schema_declared( tag = 'my_type1', tag_key = 'mytagkey' )

    class MyType2( StructValued,
      declared = d1,
      tag = 'my_type2',
      tag_key = 'mytagkey',
      struct = [ ] ):
      pass

  # tag_key mismatch
  with pytest.raises( SchemaDefinitionError ):
    d1 = schema_declared( tag = 'my_type1', tag_key = 'mytagkey2' )

    class MyType2( StructValued,
      declared = d1,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      struct = [ ] ):
      pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_values():

  class MyType1( StructValued,
    init_val = { 'x1' : True, 'x2' : False } ):

    x1 = BoolPrim()
    x2 = BoolPrim()

  assert( MyType1.schema.init_val == { 'type' : 'base', 'x1' : True, 'x2' : False } )

  expr = "$expr:py { 'x1' : True }"
  expr_normed = "$expr:py{ 'x1' : True }"

  class MyType1( StructValued,
    tag = 'my_type1',
    tag_key = 'mytagkey',
    evaluated = PJCEvaluated,
    default_val = expr,
    default_eval = { 'x1' : True, 'x2' : False } ):

    x1 = BoolPrim()
    x2 = BoolPrim()

  assert( MyType1.schema.default_val == expr_normed )
  assert( MyType1.schema.default_eval == { 'mytagkey' : 'my_type1', 'x1' : True, 'x2' : False }  )
  # this should end up using the default_eval as the init_val, since default_val is an expression
  assert( MyType1.schema.init_val == { 'mytagkey' : 'my_type1', 'x1' : True, 'x2' : False } )


  # default_eval not used if default_val is not an expression
  with pytest.raises( SchemaDefinitionError ):

    class MyType1( StructValued,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      evaluated = PJCEvaluated,
      default_val = { 'x1' : True },
      default_eval = { 'x1' : True } ):

      x1 = BoolPrim()
      x2 = BoolPrim()

  # type mismatch
  with pytest.raises( SchemaDefinitionError ):

    class MyType1( StructValued,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      evaluated = "not an evaluated",
      struct = [ ] ):
      pass

  # evaluated
  class MyType1( StructValued,
    tag = 'my_type1',
    tag_key = 'mytagkey',
    evaluated = PJCEvaluated ):

    x1 = BoolPrim()


  #.............................................................................
  v1 = MyType1.schema.decode(expr)
  v2 = v1._eval()

  print( v1, v1._encode )
  print( v2, v2._encode )

  #.............................................................................
  # default_val
  # encode evaluated
  class MyType2( StructValued,
    tag = 'my_type1',
    tag_key = 'mytagkey',
    evaluated = PJCEvaluated,
    default_val = v1 ):

    x1 = BoolPrim()

  assert( MyType2.schema.default_val == expr_normed )

  # encode valued (from evaluation)
  class MyType3( StructValued,
    tag = 'my_type1',
    tag_key = 'mytagkey',
    default_val = v2 ):

    x1 = BoolPrim()

  assert( MyType3.schema.default_val == { 'mytagkey' : 'my_type1', 'x1' : True } )

  #.............................................................................
  # default_eval

  with pytest.raises( SchemaDefinitionError ):
    # cannot be evaluated
    class MyType2( StructValued,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      default_val = expr,
      default_eval = v1,
      evaluated = PJCEvaluated ):

      x1 = BoolPrim()


  # encode valued (from evaluation, no longer evaluated)
  class MyType3( StructValued,
    tag = 'my_type1',
    tag_key = 'mytagkey',
    default_val = expr,
    default_eval = v2,
    evaluated = PJCEvaluated ):

    x1 = BoolPrim()

  assert( MyType3.schema.default_eval == { 'mytagkey' : 'my_type1', 'x1' : True } )

  with pytest.raises( SchemaDefinitionError ):
    # cannot be optional
    class MyType3( StructValued,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      default_val = expr,
      default_eval = optional,
      evaluated = PJCEvaluated ):

      x1 = BoolPrim()

  with pytest.raises( SchemaDefinitionError ):
    # cannot be an expression
    class MyType3( StructValued,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      default_val = expr,
      default_eval = expr,
      evaluated = PJCEvaluated ):

      x1 = BoolPrim()

  #.............................................................................
  # init_val

  with pytest.raises( SchemaDefinitionError ):
    # cannot be an expression

    # encode evaluated
    class MyType2( StructValued,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      init_val = v1 ):

      x1 = BoolPrim()

  # encode evaluated
  class MyType2( StructValued,
    tag = 'my_type1',
    tag_key = 'mytagkey',
    init_val = v1,
    evaluated = PJCEvaluated ):

    x1 = BoolPrim()


  # encode valued (from evaluation)
  class MyType3( StructValued,
    tag = 'my_type1',
    tag_key = 'mytagkey',
    init_val = v2 ):

    x1 = BoolPrim()

  assert( MyType3.schema.init_val == { 'mytagkey' : 'my_type1', 'x1' : True }  )

  with pytest.raises( SchemaDefinitionError ):
    # cannot be a special value
    class MyType3( StructValued,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      init_val = required ):

      x1 = BoolPrim()

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_struct():
  b1 = BoolPrim()
  b2 = BoolPrim()

  class MyType1( StructValued,
    tag = 'my_type1',
    tag_key = 'mytagkey',
    struct = [
      ('x1', b1 ),
      ('x2', b2 ) ] ):
    pass

  assert( MyType1.x1 is b1 )
  assert( MyType1.x2 is b2 )

  class MyType2( StructValued,
    tag = 'my_type2',
    tag_key = 'mytagkey',
    struct = dict(
      x1 = b1,
      x2 = b2 ) ):
    pass

  assert( MyType2.x1 is b1 )
  assert( MyType2.x2 is b2 )


  class MyType3( StructValued,
    tag = 'my_type3',
    tag_key = 'mytagkey',
    struct = odict(
      x1 = b1,
      x2 = b2 ) ):
    pass

  assert( MyType3.x1 is b1 )
  assert( MyType3.x2 is b2 )

  #.............................................................................
  # check invalid struct values

  with pytest.raises( SchemaDefinitionError ):
    # values must be schemas, not base types
    class MyType1( StructValued,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      struct = [
        ('x1', bool ),
        ('x2', bool ) ] ):
      pass

  with pytest.raises( SchemaDefinitionError ):
    # cannot be something that cannot be converted into a dict
    class MyType1( StructValued,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      struct = [1,2,3] ):
      pass

  with pytest.raises( SchemaDefinitionError ):
    # cannot be something other than dict or list
    class MyType1( StructValued,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      struct = 1 ):
      pass

  #.............................................................................
  # check invalid struct keys

  with pytest.raises( SchemaDefinitionError ):
    # must be string
    class MyType1( StructValued,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      struct = [
        ( 1, b1 ) ] ):
      pass

  with pytest.raises( SchemaDefinitionError ):
    # must be a valid python variable name
    class MyType1( StructValued,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      struct = [
        ('1', b1 ) ] ):
      pass

  with pytest.raises( SchemaDefinitionError ):
    # key cannot be a 'protected' attribute
    class MyType1( StructValued,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      struct = [
        ('_p_', b1 )] ):
      pass

  with pytest.raises( SchemaDefinitionError ):
    # key cannot be a dunder attribute
    class MyType1( StructValued,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      struct = [
        ('__', b1 )] ):
      pass

  with pytest.raises( SchemaDefinitionError ):
    # key cannot be one of the 'mapping' functions
    class MyType1( StructValued,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      struct = [
        ('items', b1 )] ):
      pass

  with pytest.raises( SchemaDefinitionError ):
    # key cannot be the same as the tag_key
    class MyType1( StructValued,
      tag = 'my_type1',
      tag_key = 'mytagkey',
      struct = [
        ('mytagkey', b1 )] ):
      pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_inheritance():

  #.............................................................................
  class MyType1( StructValued,
    tag = 'my_type1',
    tag_key = 'mytagkey1' ):
    x1 = b1

  class MyType2( StructValued,
    tag = 'my_type2',
    tag_key = 'mytagkey2' ):
    x2 = b2

  class MyType3( MyType1 ):
    pass

  assert( MyType3.tag == 'my_type1' )
  assert( MyType3.tag_key == 'mytagkey1' )
  assert( MyType3.x1 is b1 )


  #.............................................................................
  with pytest.raises( SchemaDefinitionError ):
    # no multiple inheritance
    class MyType5( MyType1, MyType2,
      tag = 'my_type5' ):
      pass


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_names():
  with pytest.raises( SchemaDefinitionError ):
    # cannot be number
    class MyType1( StructValued,
      tag = '1' ):
      pass

  with pytest.raises( SchemaDefinitionError ):
    # cannot be number
    class MyType2( StructValued,
      tag_key = '2' ):
      pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_proxy():
  b1 = BoolPrim()
  b2 = BoolPrim()


  #.............................................................................
  class MyType1( StructValued,
    tag = 'my_type1',
    tag_key = 'mytagkey1',
    struct_proxy = 'x1' ):
    x1 = b1

  class MyType2( StructValued,
    tag = 'my_type1',
    tag_key = 'mytagkey1',
    struct_proxy = 'x2' ):
    x2 = b2

  class MyType3( MyType1 ):
    pass

  assert( MyType3.schema.struct_proxy == 'x1' )

  v1 = MyType3.schema.decode(True)
  v2 = MyType3.schema.decode(False)

  assert( v1.x1 == True )
  assert( v2.x1 == False )

  #.............................................................................
  with pytest.raises( SchemaDefinitionError ):
    # struct_proxy is ambiguous
    class MyType4( MyType1, MyType2 ):
      pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_doc():

  #.............................................................................
  class MyType1( StructValued ):
    """This is MyType1
    """
    schema = dict(
      tag = 'my_type1',
      tag_key = 'mytagkey1',
      default_val = { 'x1' : True, 'x2' : True } )

    x1 = BoolPrim()
    x2 = BoolPrim()

  doc = str(MyType1.__doc__)

  print(doc)

  assert( doc.startswith("This is MyType1") )
  assert( doc.find('my_type1') > 0 )
  assert( doc.find('mytagkey1') > 0 )
  assert( doc.find('x1') > 0 )
  assert( doc.find('x2') > 0 )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_derived():
  #.............................................................................
  # derived only possible when all struct schemas have non-required default_vals
  with pytest.raises( SchemaDefinitionError ):
    class MyType1( StructValued ):
      schema = dict(
        tag = 'my_type1',
        tag_key = 'mytagkey1',
        default_val = derived )

      x1 = BoolPrim()
      x2 = BoolPrim()

    v = MyType1.schema.default_val

  class MyType2( StructValued ):
    schema = dict(
      tag = 'my_type2',
      tag_key = 'mytagkey2',
      default_val = derived )

    x1 = BoolPrim( default_val = True )
    x2 = BoolPrim( default_val = False )

  v = MyType2.schema.default_val
  assert( is_mapping(v) )

  assert( 'mytagkey2' in v )

  assert( 'x1' in v )
  assert( 'x2' in v )

  assert( v['mytagkey2'] == 'my_type2' )
  assert( v['x1'] == True )
  assert( v['x2'] == False )

  class MyType3( StructValued ):
    schema = dict(
      tag = 'my_type3',
      tag_key = 'mytagkey3',
      default_val = derived )

    y1 = MyType2

  w = MyType3.schema.default_val
  assert( is_mapping(w) )

  assert( 'mytagkey3' in w )
  assert( w['mytagkey3'] == 'my_type3' )

  assert( 'y1' in w )

  _v = w['y1']

  # should have removed the tag_key since there is no ambiguity in the schema
  assert( 'mytagkey2' not in _v )

  # should have same defaults as MyType2
  assert( 'x1' in _v )
  assert( 'x2' in _v )

  assert( _v['x1'] == True )
  assert( _v['x2'] == False )



# -*- coding: UTF-8 -*-

import io

import logging
log = logging.getLogger(__name__)

import configparser

from partis.schema_meta.base import (
  Loc,
  SchemaHint,
  SchemaError,
  is_valued_type,
  is_evaluated,
  is_string,
  is_mapping,
  is_sequence )

from .utils import (
  as_load,
  as_dump )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def dump_prep( val, level = 0 ):
  """Prepares value to be dumped, converting to standard json datastructures
  """

  if is_valued_type( val ) or is_evaluated( val ):
    val = val._encode

  if is_mapping( val ):
    if level > 1:
      raise SchemaError(
        f"Config (INI) files do not support mappings with more than two levels: {val}" )

    elif level == 0:
      _val = configparser.ConfigParser()

    else:
      _val = dict()

    for k, v in val.items():
      _val[k] = dump_prep( v, level = level + 1 )

  elif is_sequence( val ):
    raise SchemaError(
      f"Config (INI) files do not support sequence values: {val}" )

  else:
    if level <= 1:
      raise SchemaError(
        f"Config (INI) files require two levels of mapping to values: {val}" )

    _val = str(val)

  return _val

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def load_prep( val ):
  _val = dict(val)

  for k,v in _val.items():
    _val[k] = dict(v)

  if 'DEFAULT' in _val:
    _default_val = _val.pop('DEFAULT')

    for j, u in _val.items():
      for k, v in _default_val.items():
        if k not in u:
          u[k] = v

  return _val, _default_val

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def loads(
  src,
  schema = None,
  loc = None ):
  """
  Parse a Config (INI) document in a stream
  and produce the corresponding Python object.

  Note that any 'DEFAULT' section is *not* considered a part of the schema, since
  that section provides default values for all other sections.

  Parameters
  ----------
  src : str
  schema : :class:`Schema <partis.schema.struct.SchemaStruct>` | :class:`SchemaPrim <partis.schema.prim.base.SchemaPrim>`
  loc : None | :class:`Loc <partis.schema_meta.base.Loc>`
    Location information of source data (E.G. file, line/column number)

  Returns
  -------
  val : dict | :class:`Schema <partis.schema.valued.Valued>`
  default_val : dict
    Values specified in the 'DEFAULT' section.
  """

  if loc is None:
    loc = Loc()

  if isinstance( loc, str ):
    loc = Loc( filename = loc )

  try:
    parser = configparser.ConfigParser()
    parser.read_string( src )
    val, default_val = load_prep( parser )

  except BaseException as e:
    raise SchemaError(
      f"Document failed parsing",
      loc = loc,
      hints = SchemaHint.cast( e ) ) from e

  if schema is None:
    return val, default_val

  return (
    schema.schema.decode(
      val = val,
      loc = loc ),
    default_val )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def dumps( val ):
  """Serializes a suitable Python object into a Config (INI) document string

  Parameters
  ----------
  val : object | :class:`Schema <partis.schema.valued.Valued>`

  Returns
  -------
  src : str
  """

  val = dump_prep( val )

  fp = io.StringIO()

  config.write(fp)

  src = fp.getvalue()

  return src

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
load = as_load( loads )
dump = as_dump( dumps )



import pytest
import numpy as np
import array

import itertools

from partis.schema import (
  required,
  optional,
  BoolPrim,
  IntPrim,
  FloatPrim,
  StrPrim,
  SeqPrim,
  MapPrim,
  UnionPrim,
  StructValued,
  PJCEvaluated,
  SchemaError,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_declared,
  is_schema_struct_valued,
  is_valued_type,
  is_valued,
  is_evaluated,
  is_similar_value_type,
  any_schema,
  schema_declared,
  SchemaDefinitionError,
  PJCEvaluated )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
bools = [
  True,
  False ]

numerics = [
  1,
  0,
  1.0,
  0.0,
  -1.0,
  1e23 ]

strings = [
  "True",
  "False",
  "yes",
  "no",
  "1",
  "0",
  "1.0",
  "0.0",
  "\0" ]

sequences = [
  list(),
  [1,2,3],
  # TODO: fix arrays as sequences?
  # array.array('i', [1,2,3]),
  # np.array([1,2,3])
  ]

mappings = [
  dict(),
  { '1' : 1 } ]

type_combos = [
  ( is_bool, bools ),
  ( is_numeric, numerics ),
  ( is_string, strings ),
  ( is_sequence, sequences ),
  ( is_mapping, mappings ) ]

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_is_base():

  n = len(type_combos)

  for i in range(n):
    is_f = type_combos[i][0]

    for j in range(n):
      objs = type_combos[j][1]

      if i == j:
        for t in objs:
          assert( is_f(t) )
      else:
        for t in objs:
          assert( not is_f(t) )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_is_similar():
  n = len(type_combos)

  for i in range(n):
    a_objs = type_combos[i][1]

    for j in range(n):
      b_objs = type_combos[j][1]

      if i == j:
        for a in a_objs:
          for b in b_objs:
            assert( is_similar_value_type(a, b) )
      else:
        for a in a_objs:
          for b in b_objs:
            assert( not is_similar_value_type(a, b) )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_is_schema():
  s = BoolPrim()

  d = schema_declared( tag = 'my_type1' )

  class MyType1( StructValued ):
    schema = dict(
      tag = 'my_type1',
      declared = d,
      struct = [ ] )

  class MyType2( StructValued,
    tag = 'my_type2',
    struct = [ ] ):
    pass

  s1 = s.decode( val = True )
  val = MyType1( val = dict() )

  assert( is_schema_prim(s) )
  assert( not is_schema_prim(bool) )

  assert( is_schema_declared(d) )
  assert( not is_schema_declared(MyType1) )

  assert( is_schema_struct(d) )
  assert( is_schema_struct(MyType1) )

  assert( not is_schema_struct(s) )

  assert( is_schema_struct_valued(val) )
  assert( not is_schema_struct_valued(s) )
  assert( not is_schema_struct_valued(d) )
  assert( not is_schema_struct_valued(MyType1) )

  assert( is_schema(s) )
  assert( is_schema(d) )
  assert( is_schema(MyType1) )
  assert( not is_schema(bool) )

  assert( any_schema(
    val = s,
    schemas = [ s, ] ) )

  assert( any_schema(
    val = s1,
    schemas = [ s, ] ) )

  assert( any_schema(
    val = val,
    schemas = [ MyType1, ] ) )

  assert( not any_schema(
    val = val,
    schemas = [ MyType2, ] ) )

  assert( any_schema(
    val = MyType1,
    schemas = [ MyType1, ] ) )

  assert( not any_schema(
    val = MyType1,
    schemas = [ MyType2, ] ) )

  assert( not any_schema(
    val = bool,
    schemas = [ s, ] ) )

  assert( not any_schema(
    val = bool,
    schemas = [ bool, ] ) )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_is_valued():
  s = BoolPrim(
    evaluated = PJCEvaluated )

  class MyType1( StructValued,
    tag = 'my_type1',
    struct = [ ] ):
    pass

  s1 = s.decode( val = True )
  s2 = s.decode( val = "$expr:py True" )
  val1 = MyType1( val = dict() )

  assert( is_valued_type(s1) )
  assert( is_valued_type(val1) )
  assert( not is_valued_type(True) )

  assert( is_valued(s1) )
  assert( is_valued([ s1, s1, s1 ]) )
  assert( is_valued({ 'a': s1, 'b': s1, 'c': s1 }) )
  assert( is_valued(val1) )
  assert( is_valued(True) )

  assert( not is_valued(s2) )
  assert( not is_valued([ s2, s1, s1 ]) )
  assert( not is_valued({ 'a': s2, 'b': s1, 'c': s1 }) )

  assert( is_evaluated(s2._src) )
  assert( not is_evaluated(s1._src) )
  assert( not is_valued_type(s2._src) )



import re
import logging
log = logging.getLogger(__name__)

from collections.abc import (
  Mapping,
  Sequence )

from numbers import (
  Number,
  Complex,
  Real,
  Integral )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
from partis.utils.special import (
  SpecialType,
  RequiredType,
  OptionalType,
  DerivedType,
  NotSetType )

from partis.utils import (
  Loc,
  ModelHint,
  ModelError )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import partis


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
name_re = r"^[a-zA-Z\_][a-zA-Z\_0-9]*$"
name_cre = re.compile( name_re )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaError( ModelError ):
  """Base of all schema related errors
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaNameError( SchemaError ):
  """Raised when a name error occurs
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaParseError( SchemaError ):
  """Raised when a schema cannot be detected
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaDetectionError( SchemaError ):
  """Raised when a schema cannot be detected
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaDeclaredError( SchemaError ):
  """Raised when a schema's declared is used before the schema has been defined
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaDefinitionError( SchemaError ):
  """Raised when a schema's definition is not correctly specified
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaValidationError( SchemaError ):
  """Raised when data could not be validated (encoded/decoded) using schema definition
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaEvaluationError( SchemaError ):
  """Raised when evaluated expression failed to result in a decoded value
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaHint( ModelHint ):
  """Base of all schema hints
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def assert_valid_name( name ):
  if not isinstance( name, str ):
    raise SchemaNameError(
      f"Name must be a string: {name}")

  match = name_cre.fullmatch(name)

  if match is None:
    raise SchemaNameError(
      f"Name must be only alpha-numeric or underscrores, and not start with a number: {name}")


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def assert_valid_path( path ):

  if isinstance( path, str ):
    if path == '':
      # empty path is valid
      return

    path = path.split('.')

  if not isinstance( path, list ) and all( isinstance(p, str) for p in path ):
    raise SchemaNameError(
      f"Name path must be a string or list of strings: {path}")

  for i, name in enumerate(path):
    try:

      assert_valid_name( name )

    except SchemaNameError as e:
      raise SchemaNameError(
        f"Path name segment {i} not valid: {path}",
        hints = SchemaHint.cast( e ) ) from e


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_special( val ):
  return isinstance( val, SpecialType )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_required( val ):
  """
  Returns
  -------
  bool
    True if the value corresponds to being required, likely resulting in
    an error if no value is supplied.
  """

  return isinstance( val, RequiredType )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_optional( val ):
  """
  Returns
  -------
  bool
    True if the value corresponds to being optional, resulting in a default of None
  """
  return val is None or isinstance( val, OptionalType )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_derived( val ):
  """
  Returns
  -------
  bool
    True if the value is to be derived from some other value(s)
  """
  return isinstance( val, DerivedType )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_notset( val ):
  """
  Returns
  -------
  bool
    True if the value corresponds to being not set, or otherwise undefined,
    including not even ``None``.
  """

  return isinstance( val, NotSetType )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_bool( obj ):
  """Is a boolean value
  """

  if isinstance( obj, type ):
    return issubclass( obj, bool ) or issubclass( obj, partis.schema.valued.BoolValued )

  return isinstance( obj, bool ) or isinstance( obj, partis.schema.valued.BoolValued )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_numeric( obj ):
  """Is a numeric value
  """

  if isinstance( obj, type ):
    return issubclass( obj, Number ) and not is_bool( obj )

  return isinstance( obj, Number ) and not is_bool( obj )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_string( obj ):
  """Is a string value
  """

  if isinstance( obj, type ):
    return issubclass( obj, str )

  return isinstance( obj, str )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_sequence( obj ):
  """Is a sequence value, but not a string
  """
  if isinstance( obj, type ):
    return issubclass( obj, Sequence ) and not is_string( obj )

  return isinstance( obj, Sequence ) and not is_string( obj )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_mapping( obj ):
  """Is a mapping value
  """
  if isinstance( obj, type ):
    return issubclass( obj, Mapping )

  return isinstance( obj, Mapping )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
type_tests = [
  is_required,
  is_derived,
  is_optional,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping ]

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_similar_value_type( a, b ):
  """
  Returns
  -------
  is_similar_value_type : bool
    True if all `type_tests` return the same for both values
  """
  return all( f(a) == f(b) for f in type_tests )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_schema( obj ):
  """Is a schema class or schema declared
  """
  return issubclass( type(obj), partis.schema_meta.schema.SchemaRef )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_schema_prim( obj ):
  """Is a schema primitive type
  """
  return issubclass( type(type(obj)), partis.schema_meta.prim.SchemaPrimMeta )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_schema_declared( obj ):
  """Is a schema class or schema declared type
  """
  return issubclass( type(obj), partis.schema_meta.schema.SchemaDeclared )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_schema_struct( obj ):
  """Is a schema class or schema declared type
  """
  return (
    issubclass( type(obj), partis.schema_meta.struct.SchemaStruct )
    or issubclass( type(obj), partis.schema_meta.struct.SchemaStructDeclared )
    or issubclass( type(obj), partis.schema_meta.struct.SchemaStructProxy ) )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_schema_struct_valued( obj ):
  """Is an instance value of a schema class type
  """
  return (
    is_valued_type( obj )
    and is_schema_struct(obj._schema) )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_evaluated_class( val ):
  """Is an evaluated class type
  """
  return ( issubclass( type(val), partis.schema_meta.eval.EvaluatedMeta ) )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_provider( val ):
  """Is an evaluated type
  """
  return issubclass( type(type(val)), partis.schema_meta.eval.ProviderMeta )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_evaluated( val ):
  """Is an evaluated type
  """
  return issubclass( type(type(val)), partis.schema_meta.eval.EvaluatedMeta )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_valued_type( val ):
  """
  Returns
  -------
  is_valued_type : bool
    True if is an instance value of any schema type
  """
  return issubclass( type(type(val)), partis.schema_meta.valued.ValuedMeta )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def any_schema( val, schemas ):
  """Is the schema any of the list schemas

  Parameters
  ----------
  schema : object
    Schema to test
  schema : list[object]
    Schemas to compare to
  """
  if is_valued_type( val ):
    return any( val._schema is s.schema for s in schemas )

  elif is_schema( val ):
    return any( val.schema is s.schema for s in schemas )

  return False


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_valued( val ):
  """
  Returns
  -------
  is_valued : bool
    True if value is completely evaluated
  """

  if is_evaluated(val):
    return False

  if is_valued_type( val ):
    return val._valued

  if is_sequence( val ):
    return all( is_valued(v) for v in val )

  if is_mapping( val ):
    return all( is_valued(v) for v in val.values() )

  return True



# -*- coding: UTF-8 -*-

import logging
log = logging.getLogger(__name__)

from copy import copy
import inspect
import re

from partis.utils import (
  odict,
  fmt_base_or_type,
  fmt_iterable_or,
  fmt_attr_doc )

from partis.schema_meta.base import (
  SchemaError,
  SchemaNameError,
  SchemaDeclaredError,
  SchemaDefinitionError,
  SchemaValidationError,
  SchemaHint,
  Loc,
  assert_valid_name,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_declared,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_evaluated,
  is_valued,
  is_optional,
  is_required )

from partis.schema.valued import (
  SeqValued )

from . import (
  SchemaPrimDeclared,
  SchemaPrim )

from partis.schema_meta.schema import (
  fmt_schema_typename )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SeqPrimDeclared( SchemaPrimDeclared ):
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SeqPrim( SchemaPrim ):
  """Primitive for list/sequenced values

  Parameters
  ----------
  item : :class:`SchemaPrim <partis.schema.struct.SchemaStruct>` | :class:`SchemaPrim <partis.schema.prim.base.SchemaPrim>`
    The schema for each item in the list
  min_len : int | None
  max_len : int | None
  **kwargs :
    arguments passed to :class:`SchemaPrim <partis.schema.prim.SchemaPrim>`
  """

  #-----------------------------------------------------------------------------
  def __new__( mcls,
    name,
    bases,
    namespace,
    item,
    min_len = None,
    max_len = None,
    **kwargs ):

    loc = kwargs.get('loc', None)

    if not is_schema( item ):
      raise SchemaDefinitionError(
        f"`SeqPrim` item must be instance of `SchemaPrim` or `Schema`: {item}",
        loc = loc )

    if min_len is None:
      min_len = 0

    min_len = int(min_len)

    if min_len < 0:
      raise SchemaDefinitionError(
        f"`SeqPrim` minimum length must be >= 0: {min_len}",
        loc = loc )

    if max_len is not None:
      max_len = int(max_len)

      if max_len < min_len:
        raise SchemaDefinitionError(
          f"`SeqPrim` maximum length must be >= {min_len}: {max_len}",
          loc = loc )

    namespace['_p_item'] = item
    namespace['_p_min_len'] = min_len
    namespace['_p_max_len'] = max_len

    if not item.schema_defined:
      kwargs['schema_deps'] = kwargs.get('schema_deps', list()) + [item]

    return super().__new__( mcls,
      name = name,
      bases = bases,
      namespace = namespace,
      valued_type = SeqValued,
      **kwargs )

  #-----------------------------------------------------------------------------
  def __init__( cls,
    name,
    bases,
    namespace,
    item,
    min_len = None,
    max_len = None,
    **kwargs ):


    super().__init__(
      name = name,
      bases = bases,
      namespace = namespace,
      **kwargs )

  #-----------------------------------------------------------------------------
  def _get_attr_doc_lines( cls,
    noindex = False,
    depth = 0,
    max_depth = 2 ):

    lines = super()._get_attr_doc_lines(
      noindex = noindex,
      depth = depth,
      max_depth = max_depth )

    lines.extend([
      fmt_attr_doc(
        name = attr,
        typename = type( getattr( cls, attr ) ),
        val = getattr( cls, attr ),
        prefix = 'schema',
        noindex = noindex )
      for attr in [
        'min_len',
        'max_len' ] ])


    lines.append(
      fmt_attr_doc(
        name = 'item',
        typename = fmt_schema_typename( cls.item ),
        val = ...,
        prefix = 'schema',
        noindex = noindex,
        doc = cls.item._get_doc(
          # always specify noindex since the item schema already documented
          noindex = True,
          depth = depth + 1,
          max_depth = max_depth ) ) )

    return lines

  #-----------------------------------------------------------------------------
  def _schema_hash_children( cls ):
    return [ cls.item, ]


  #-----------------------------------------------------------------------------
  @property
  def hints( cls ):
    hints = super().hints

    hints.append( f"min_len: {cls.min_len}" )
    hints.append( f"max_len: {cls.max_len}" )

    if is_schema_declared( cls.item ):
      hints.append( SchemaHint(
        f"item: {cls.item.tag}") )

    else:
      hints.append( SchemaHint(
        f"item:",
        hints = cls.item.schema.hints ) )

    return hints

  #-----------------------------------------------------------------------------
  @property
  def item( cls ):
    return cls._p_item

  #-----------------------------------------------------------------------------
  @property
  def min_len( cls ):
    return cls._p_min_len

  #-----------------------------------------------------------------------------
  @property
  def max_len( cls ):
    return cls._p_max_len

  #-----------------------------------------------------------------------------
  def validate( cls,
    val,
    loc = None ):

    if not is_sequence( val ):
      raise SchemaValidationError(
        f"Must be a `sequence`: {type(val).__name__}",
        loc = loc )

    if len(val) < cls.min_len:
      raise SchemaValidationError(
        f"Must have length >= {cls.min_len}: {len(val)}",
        loc = loc )

    if cls.max_len is not None and len(val) > cls.max_len:
      raise SchemaValidationError(
        f"Must have length <= {cls.max_len}: {len(val)}",
        loc = loc )


    return super().validate(
      val = val,
      loc = loc )

  #-----------------------------------------------------------------------------
  def encode( cls,
    val = None,
    loc = None ):

    val, loc, validate = cls._encode(
      val = val,
      loc = loc )

    if not validate:
      return val

    val = cls.validate(
      val = val,
      loc = loc )

    val = [ cls.item.schema.encode(v) for v in val ]

    if is_schema_struct( cls.item ):
      tag_key = cls.item.tag_key

      for v in val:
        if is_mapping( v ):
          v.pop( tag_key )

    return val


  #-----------------------------------------------------------------------------
  def decode( cls, val, loc = None ):

    val, loc, validate = cls._decode(
      val = val,
      loc = loc )

    if not validate:
      return val

    val = cls.validate(
      val = val,
      loc = loc )

    decode_val = list()

    for i, _val in enumerate(val):
      _loc = loc( val, i )

      try:
        _val = cls.item.schema.decode(
          val = _val,
          loc = _loc )

      except SchemaError as e:
        raise SchemaValidationError(
          f"List item invalid: {i}",
          loc = _loc,
          hints = SchemaHint.cast( e ) ) from e

      decode_val.append( _val )

    return cls.valued_type(
      val = decode_val,
      schema = cls,
      loc = loc )




from copy import copy

import re
import uuid
import sys

from abc import ABCMeta


import logging
log = logging.getLogger(__name__)

from partis.utils import (
  adict,
  odict,
  rdict,
  rlist,
  StringFunction,
  mapping_attrs,
  protected_attr,
  indent_lines,
  fmt_base_or_type,
  fmt_class_name,
  make_dynamic_class_name )


from .base import (
  SchemaError,
  SchemaNameError,
  SchemaDeclaredError,
  SchemaDefinitionError,
  SchemaValidationError,
  SchemaHint,
  Loc,
  assert_valid_name,
  assert_valid_path,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_optional,
  is_required,
  is_derived,
  is_similar_value_type,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_evaluated_class,
  is_evaluated,
  is_valued,
  is_valued_type,
  any_schema )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaPrimMeta( type ):
  """Meta-class for schema primitives

  .. note::

    The primary purpose of this metaclass is to instantiate primitive schema
    types without the typical class bodies, and providing convenience to generating
    class names.
  """

  #-----------------------------------------------------------------------------
  def __call__( pcls,
    name = None,
    module = None,
    module_set = None,
    doc = None,
    **kwargs ):
    """Creates a new primitive schema class of the primitive meta type

    Parameters
    ----------
    name : str
      Name of sub-class. If not given, a name will be generated automatically
    module : None | str | ModuleType
      Module to use for the subclass.
      If None, the module will be determined from the caller frame.
      If a string, the module must be importable.
    module_set : None | bool
      If True (or None), adds the subclass to the module `module`.
      If False, the return value will be the only reference to the subclass.
    **kwargs :
      Arguments passed to the class __new__ and __init__

    Note
    ----
    This modifies the call signature using the prim class as a generator function
    """


    module, name = make_dynamic_class_name(
      default_name = pcls.__name__,
      module = module,
      name = name )

    bases = tuple()

    namespace = pcls.__prepare__(
      name,
      bases,
      **kwargs )

    namespace['__module__'] = module.__name__

    namespace['__doc__'] = doc

    cls = super().__call__(
      # pcls,
      name,
      bases,
      namespace,
      **kwargs )

    if module_set is None or module_set:

      if hasattr( module, cls.__name__ ):
        log.warning(f"Overriding module attribute with class definition: {cls.__name__}")

      setattr( module, cls.__name__, cls )

    return cls



import pytest
from pytest import raises

from pprint import pprint

import itertools

from partis.schema import (
  required,
  optional,
  BoolPrim,
  IntPrim,
  FloatPrim,
  StrPrim,
  SeqPrim,
  MapPrim,
  UnionPrim,
  SchemaStruct,
  PJCEvaluated,
  NotEvaluated,
  SchemaError,
  SchemaDefinitionError,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  any_schema, )

from partis.schema.serialize.yaml import (
  loads,
  dumps )

from test_prim import (
  is_defined,
  get_val,
  should_pass,
  max_tests,
  step_tests )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# use this to only run sub-set of all possible tests
# max_tests = -1
# step_tests = 1

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def prim_str_cases( ):
  default_vals = [
    optional,
    required,
    "xyz",
    "asd  asdd" ]

  restricteds = [
    None,
    ["qwe"],
    ["xyz"],
    ["qwe", "asd  asdd"] ]

  evaluateds = [
    None,
    PJCEvaluated ]

  vals = [
    None,
    "$expr:py self",
    "$func:py return self",
    "xyz",
    "qwe",
    "asd  asdd",
    "abc",
    "123.5" ]

  test_cases = itertools.product(
    vals,
    default_vals,
    restricteds,
    evaluateds )

  if max_tests >= 0:
    test_cases = itertools.islice( test_cases, 0, max_tests * step_tests, step_tests )

  return test_cases


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
@pytest.fixture( params = prim_str_cases() )
def prim_str_case( request ):
  return request.param

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_prim_str( prim_str_case ):

  val, default_val, restricted, evaluated = prim_str_case


  if should_pass(
    val = val,
    is_type = is_string,
    default_val = default_val,
    restricted = restricted,
    evaluated = evaluated ):

    p = StrPrim(
      default_val = default_val,
      restricted = restricted,
      evaluated = evaluated,
      doc = "My string" )

    assert p.default_val == default_val
    assert p.restricted == restricted
    assert p.doc.startswith("My string")
    assert p.evaluated == NotEvaluated if evaluated is None else evaluated

    _val = p.decode( val )

    if evaluated is not None and evaluated.check(val):
      if is_defined(default_val):
        # only check evaluated when default value is defined
        _val = _val._eval( context = { 'self' : default_val } )
        assert _val == p.default_val

    else:
      assert _val == get_val( val, default_val )

  else:
    with raises(SchemaError):

      p = StrPrim(
        default_val = default_val,
        restricted = restricted,
        evaluated = evaluated,
        doc = "My string" )

      _val = p.decode( val )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_prim_str_pattern():
  p = StrPrim(
    pattern = r"\w\d\w\d",
    doc = "My string" )

  x = p.decode( "a1b2" )
  assert x == "a1b2"

  with raises(SchemaError):
    x = p.decode( "a1b2c3" )

  with raises(SchemaError):
    x = p.decode( "abc" )

  with raises(SchemaDefinitionError):
    p = StrPrim(
      default_val = "a1b2c3",
      pattern = r"\w\d\w\d",
      doc = "My string" )

  with raises(SchemaDefinitionError):
    p = StrPrim(
      restricted = [ "a1b2c3" ],
      pattern = r"\w\d\w\d",
      doc = "My string" )

  with raises(SchemaDefinitionError):
    p = StrPrim(
      pattern = r"asd(&*^$*&G##&])",
      doc = "My string" )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_prim_str_char_case():
  with raises(SchemaError):
    p = StrPrim(
      char_case = 'asdasd',
      doc = "My string" )

  s_upper = "A1B2C3"
  s_lower = "a1b2c3"

  p_upper = StrPrim(
    default_val = s_lower,
    char_case = 'upper',
    doc = "My string" )

  p_lower = StrPrim(
    default_val = s_upper,
    char_case = 'lower',
    doc = "My string" )

  assert p_upper.default_val == s_upper
  assert p_upper.decode() == s_upper
  assert p_upper.decode(s_upper) == s_upper
  assert p_upper.decode(s_lower) == s_upper

  assert p_upper.encode() is None
  assert p_upper.encode(s_upper) == s_upper
  assert p_upper.encode(s_lower) == s_upper

  assert p_lower.default_val == s_lower
  assert p_lower.decode() == s_lower
  assert p_lower.decode(s_upper) == s_lower
  assert p_lower.decode(s_lower) == s_lower

  assert p_lower.encode() is None
  assert p_lower.encode(s_upper) == s_lower
  assert p_lower.encode(s_lower) == s_lower

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_prim_str_strip():
  val = "asd"
  sval = f"  {val}  "

  p = StrPrim(
    strip = True,
    default_val = sval )

  assert p.default_val == val
  assert p.decode() == val
  assert p.decode(val) == val
  assert p.decode(sval) == val

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_prim_nonempty_str():

  with raises(SchemaDefinitionError):
    p = StrPrim(
      nonempty = True,
      default_val = required )

  with raises(SchemaDefinitionError):
    p = StrPrim(
      nonempty = True,
      default_val = optional )

  with raises(SchemaDefinitionError):
    p = StrPrim(
      nonempty = True,
      default_val = "" )

  p = StrPrim(
    nonempty = True,
    default_val = "asd" )

  p = StrPrim(
    nonempty = True,
    init_val = "asd",
    default_val = required )

  p = StrPrim(
    nonempty = True,
    init_val = "asd",
    default_val = optional )

  assert p.decode() == None

  with raises(SchemaError):
    p.decode("")




from .valued import Valued

from .bool_valued import BoolValued
from .int_valued import IntValued
from .float_valued import FloatValued
from .str_valued import StrValued
from .seq_valued import SeqValued
from .map_valued import MapValued
from .struct_valued import StructValued



"""Base elements for defining a data schema, including base types, data structures,
and data classes (:class:`SchemaStruct <partis.schema.struct.SchemaStruct>`)
"""

from partis.utils.special import (
  required,
  optional,
  derived,
  notset )

from .module import (
  SchemaLoader,
  SchemaFinder,
  SchemaModule )

from .eval import (
  ProviderSupport,
  ProviderMeta,
  Provider,
  EvaluatedMeta,
  Evaluated,
  NotEvaluated,
  EvaluatedContext,
  ConstEvaluatedContext,
  PyEvaluated,
  CheetahEvaluated,
  PJCEvaluated,
  EvalFunc )

from .valued import (
  BoolValued,
  IntValued,
  FloatValued,
  StrValued,
  SeqValued,
  MapValued,
  StructValued )

from .prim import (
  PassPrim,
  BoolPrimDeclared,
  BoolPrim,
  IntPrimDeclared,
  IntPrim,
  FloatPrimDeclared,
  FloatPrim,
  StrPrimDeclared,
  StrPrim,
  SeqPrimDeclared,
  SeqPrim,
  MapPrimDeclared,
  MapPrim,
  UnionPrimDeclared,
  UnionPrim )

from .declared import (
  schema_declared )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# import everything from `schema_meta` to have unified "external" imports
# NOTE: `schema_meta` is not meant to be exposed in the main package because it's
# designed to lazily import the `meta` components to avoid issues with circular
# imports while initializing the various modules.
from partis.schema_meta.base import (
  assert_valid_name,
  SchemaError,
  SchemaParseError,
  SchemaDetectionError,
  SchemaDeclaredError,
  SchemaDefinitionError,
  SchemaValidationError,
  SchemaValidationError,
  SchemaEvaluationError,
  SchemaHint,
  Loc,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_special,
  is_optional,
  is_required,
  is_derived,
  is_similar_value_type,
  is_schema_prim,
  is_schema_declared,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_provider,
  is_evaluated_class,
  is_evaluated,
  is_valued,
  is_valued_type,
  any_schema )

from partis.schema_meta.valued import (
  ValuedMeta )

from partis.schema_meta.prim import (
  SchemaPrimMeta )

from partis.schema_meta.property import (
  SchemaProperty,
  ConstProperty )

from partis.schema_meta.schema import (
  PassValued,
  SchemaRef,
  SchemaProxy,
  SchemaDep,
  SchemaDeclared,
  Schema )

from partis.schema_meta.struct import (
  SchemaStructDeclared,
  SchemaStruct,
  SchemaStructProxy )

from . import serialize

from .hint import (
  Hint,
  HintList )



import pytest

from partis.schema import (
  SchemaError,
  Hint )

from partis.schema.serialize.yaml import (
  loads,
  dumps )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_hint():
  for lvl in [ "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL" ]:

    hint = Hint(
      msg = "test message",
      loc = "from somewhere",
      level = lvl,
      hints = [
        dict(
          msg = "another message",
          loc = "from somewhere else",
          level = lvl ),
        None ] )

    _hint = hint._cast()
    assert( _hint.msg == "test message" )
    assert( _hint.loc == "from somewhere" )
    assert( _hint.level == lvl )
    assert( len(_hint.hints) == 1 )
    assert( _hint.hints[0].msg == "another message" )
    assert( _hint.hints[0].loc == "from somewhere else" )
    assert( _hint.hints[0].level == lvl )

    s = _hint.fmt()

    assert( s.startswith( "test message" ) )




from copy import copy

from collections.abc import (
  Mapping )

import logging
log = logging.getLogger(__name__)

from partis.utils import (
  adict,
  odict )

from partis.schema_meta.base import (
  SchemaError,
  SchemaNameError,
  SchemaDeclaredError,
  SchemaDefinitionError,
  SchemaValidationError,
  SchemaHint,
  Loc,
  assert_valid_name,
  is_notset,
  is_valued,
  is_evaluated,
  is_mapping,
  is_derived,
  is_optional,
  is_required,
  is_valued_type,
  is_schema_declared,
  is_schema_struct )

from partis.schema_meta.struct import (
  SchemaStructProxy )

from partis.schema_meta.valued import (
  ValuedMeta )

from partis.schema import (
  EvaluatedContext )

from .valued import (
  Valued,
  get_schema_loc,
  get_init_val,
  get_src_val )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class StructValuedMeta( ValuedMeta, SchemaStructProxy ):
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def extract_val_loc( args, kwargs ):

  if len(args) == 0:
    val = None
    loc = None
  else:
    if len(args) == 1:
      val = args[0]
      loc = None
    elif len(args) == 2:
      val, loc = args
    else:
      raise ValueError(
        f"positional arguments must be at most `(val, loc)`."
        " All keyword arguments are interpreted as items of `val`")

  if kwargs:
    if val:
      val = {**val, **kwargs}
    else:
      val = kwargs

  return val, loc

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class StructValued( Valued, Mapping, metaclass = StructValuedMeta ):
  """Container for fully or partially evaluated fixed mapping data.

  .. note::

    The terminology of 'struct' is used to indicate that the allowed key-value
    pairs of the value are fixed, such that keys may not be added or removed
    dynamically.
    However, key-values may be heterogenous, with a unique schema defined for
    each key, unlike the generic MapValue that specifies the same schema for
    all keys.

  Parameters
  ----------
  val : NoneType | object
    Source data used to initialize values
  loc : NoneType | :class:`Loc <partis.schema_meta.base.Loc>`
    Location information of source data (E.G. file, line/column number)

  Raises
  ------
  SchemaValidationError
    If the value is not valid
  """

  #-----------------------------------------------------------------------------
  def __new__( cls,
    *args, **kwargs ):

    val, loc = extract_val_loc( args, kwargs )

    if is_valued_type( val ) or is_evaluated( val ):
      if loc is None:
        loc = val._loc

      # NOTE: must still validate value since there are no gaurantees about
      # the original schema
      val = val._encode

    if loc is None:
      loc = Loc()

    if not isinstance( loc, Loc ):
      raise SchemaValidationError(
        f"`loc` must be instance of `Loc`: {type(loc)}")

    if is_optional(val):
      val = cls.schema.default_val

      if is_required( val ):
        raise SchemaValidationError(
          f"Value is required",
          loc = loc,
          hints = cls.schema.hints )

      elif is_optional( val ):
        return None

    valued = True

    if cls.schema.evaluated.check( val ):
      valued = False

      val = cls.schema.evaluated(
        schema = cls.schema,
        src = val,
        loc = loc )

    self = super().__new__( cls )

    self._p_dict = odict()
    self._p_loc = loc
    self._p_valued = valued
    self._p_src = val

    return self

  #-----------------------------------------------------------------------------
  def __init__( self,
    *args, **kwargs ):

    val, loc = extract_val_loc( args, kwargs )

    # these are set by the __new__ method
    if not self._valued:
      return

    val = self._src
    loc = self._loc

    tag_key = self._schema.tag_key
    tag = self._schema.tag

    self._p_dict[tag_key] = tag

    loc = self._loc

    if not is_mapping( val ):
      if self._schema.struct_proxy:
        # uses source data for primary struct key value with remaining defaults

        _val = dict()
        _val[ self._schema.struct_proxy ] = val
        val = _val

      else:
        raise SchemaValidationError(
          f"`val` must be a mapping: {val}",
          loc = loc )

    all_keys = set(self._schema.struct.keys()) | set(val.keys())

    _locs = {
      k : loc( val, k )
      for k in all_keys }

    if tag_key in val:
      _tag = val[tag_key]

      if _tag != tag:
        log.warning(f"schema {self._schema.__module__}.{self._schema.__name__} expected `{tag_key}` of `{tag}`: `{_tag}`, {loc}")

    _keys = list(self._schema.struct.keys())

    for k in val.keys():

      if k not in _keys and k != tag_key:
        # issue warnings for unexpected keys
        log.warning(f"not in schema {self._schema.__module__}.{self._schema.__name__}: '{k}', {_locs[k]}")


    for k, _schema in self._schema.struct.items():
      _schema = _schema.schema

      # decode data according to schema
      _val = None
      _loc = _locs[k]

      if k in val:
        _val = val[k]

      try:
        _val = _schema.decode(
          val = _val,
          loc = _loc )

      except SchemaError as e:

        raise SchemaValidationError(
          f"Schema item `{k}` could not be decoded",
          loc = _loc,
          hints = SchemaHint.cast( e ) ) from e

      if _val is not None and not is_valued(_val):
        self._p_valued = False

      self._p_dict[k] = _val

  #-----------------------------------------------------------------------------
  @property
  def _valued( self ):
    """bool : False if any item is not fully evaluated. True otherwise.
    """
    return self._p_valued

  #-----------------------------------------------------------------------------
  @property
  def _src( self ):
    return self._p_src

  #-----------------------------------------------------------------------------
  @property
  def _loc( self ):
    """:class:`Loc <partis.schema_meta.base.Loc>` : Location information of source
    data (E.G. file, line/column number)
    """
    return self._p_loc

  #-----------------------------------------------------------------------------
  @property
  def _schema( self ):
    """:class:`SchemaStruct <partis.schema.struct.SchemaStruct>` :  Schema class
    of this instance
    """
    return type(self).schema

  #-----------------------------------------------------------------------------
  @property
  def _encode( self ):
    """dict : Plain-data values of items.
    """
    if is_evaluated( self._src ):
      return self._src._encode

    val = odict()

    val[ self._schema.tag_key ] = self._schema.tag

    for k, _schema in self._schema.struct.items():
      _schema = _schema.schema

      _val = self._p_dict[k]

      if is_valued_type( _val ) or is_evaluated( _val ):
        _val = _val._encode

      if _val is not None:
        if is_mapping( _val ) and is_schema_struct( _schema ):
          # don't need to store tag key since schema is not ambiguous
          _val.pop( _schema.tag_key )

        val[k] = _val

    return val


  #-----------------------------------------------------------------------------
  def __str__( self ):
    # return self._p_dict.__str__( )
    return f"{type(self).__name__}({list(self._p_dict.items())})"

  #-----------------------------------------------------------------------------
  def __repr__( self ):
    # return self._p_dict.__repr__( )
    return f"{type(self).__name__}({list(self._p_dict.items())})"


  #-----------------------------------------------------------------------------
  def __len__( self ):
    return len(self._p_dict)

  #-----------------------------------------------------------------------------
  def __iter__( self ):
    return iter(self._p_dict)

  #-----------------------------------------------------------------------------
  def keys( self ):
    return self._p_dict.keys()

  #-----------------------------------------------------------------------------
  def values( self ):
    return self._p_dict.values()

  #-----------------------------------------------------------------------------
  def items( self ):
    return self._p_dict.items()

  #-----------------------------------------------------------------------------
  def get( self, key, default = None ):
    return self._p_dict.get( key, default )


  # #-----------------------------------------------------------------------------
  # def __getattribute__( self, name ):
  #
  #   try:
  #     return super().__getattribute__(name)
  #
  #   except AttributeError as e:
  #
  #     if name in self._p_dict:
  #       return self._p_dict[name]
  #
  #     raise AttributeError(
  #       f"'{type(self).__name__}' object has no attribute '{name}'") from e
  #
  # #-----------------------------------------------------------------------------
  # def __setattr__( self, name, val ):
  #   if name in self._schema.struct or name == self._schema.tag_key:
  #     self[name] = val
  #     return
  #
  #   if not name.startswith('_'):
  #     raise AttributeError(f"Cannot assign new exposed attribute: {name}")
  #
  #   super().__setattr__( name, val )

  #-----------------------------------------------------------------------------
  def __iter__( self ):
    return iter(self._p_dict)

  #-----------------------------------------------------------------------------
  def __copy__( self ):

    obj = type(self)( self._p_dict, self._loc )

    return obj

  #-----------------------------------------------------------------------------
  def __getitem__( self, key ):
    if key == self._schema.tag_key:
      return self._schema.tag

    if key not in self._schema.struct:
      raise KeyError(
        f"Schema does not contain key: {key}")

    return self._p_dict[key]

  #-----------------------------------------------------------------------------
  def __setitem__( self, key, val ):

    loc = None

    if is_valued_type( val ):
      loc = val._loc

    tag_key = self._schema.tag_key

    if key == tag_key:
      if val != self._schema.tag:
        raise SchemaValidationError(f"Schema value for`{tag_key}` must be `{self._schema.tag}`: {val}")

    else:

      if key not in self._schema.struct:
        raise SchemaValidationError(f"Schema does not contain key: {key}")

      try:
        val = self._schema.struct[key].schema.decode(
          val = val,
          loc = loc )

      except SchemaError as e:
        raise SchemaValidationError(
          f"value error for key `{key}`",
          hints = [
            SchemaHint("value", hints = val),
            SchemaHint.cast( e ) ] ) from e


      self._p_dict[ key ] = val


  #-----------------------------------------------------------------------------
  def _lint( self,
    context = None,
    logger = None ):
    """Lints source

    Parameters
    ----------
    context : None | :class:`EvaluatedContext` | list[ :class:`EvaluatedContext` ]
      Sets the context for which this expression is to be evaluated.
      The expression is evaluated only when the specified context is equivalent
      to the one for this evaluated class.
    logger : logging.Logger
      A logger object to direct expression logs/prints instead of stdout

    Returns
    -------
    : list[SchemaHint]

    See Also
    --------
    :class:`Evaluated <partis.schema.eval.Evaluated>`
    """

    hints = list()

    if self._valued:
      return hints

    # must evaluate to get value

    if is_evaluated( self._p_src ):

      hints.extend( self._p_src._lint(
        context = context,
        logger = logger ) )

      return hints

    # local variables for items in list or dictionary
    # NOTE: parent of items is the current value, parent of current will be
    # parent of the item parent

    for k, schema in self._schema.struct.items():
      v = self[k]

      if not is_valued(v):

        if isinstance( context, EvaluatedContext ):
          _context = context(
            schema = self._schema,
            parent = self,
            key = k )

        else:
          _context = context

        hints.extend(  v._lint(
          context = _context,
          logger = logger ) )

    if hints:
      return [
        SchemaHint(
          loc = Loc(
            path = self._loc.path,
            line = self._loc.line,
            col = self._loc.col ),
          hints = hints ) ]

    return hints

  #-----------------------------------------------------------------------------
  def _eval( self,
    context = None,
    logger = None ):
    """Evaluates source

    Parameters
    ----------
    context : None | :class:`EvaluatedContext` | list[ :class:`EvaluatedContext` ]
      Sets the context for which this expression is to be evaluated.
      The expression is evaluated only when the specified context is equivalent
      to the one for this evaluated class.
    logger : logging.Logger
      A logger object to direct expression logs/prints instead of stdout

    Returns
    -------
    value : :class:`SchemaStruct <partis.schema.struct.SchemaStruct>`
      Decoded value resulting from evaluation.
      If there are no un-evaluated values, then will return a shallow copy of
      the current data.

    See Also
    --------
    :class:`Evaluated <partis.schema.eval.Evaluated>`
    """

    if self._valued:
      return copy(self)

    # must evaluate to get value

    if is_evaluated( self._p_src ):

      return self._p_src._eval(
        context = context,
        logger = logger )

    # local variables for items in list or dictionary
    # NOTE: parent of items is the current value, parent of current will be
    # parent of the item parent
    val = adict()

    for k, schema in self._schema.struct.items():
      v = self[k]

      if is_valued(v):
        val[k] = v

      else:
        if isinstance( context, EvaluatedContext ):
          _context = context(
            schema = self._schema,
            parent = val,
            key = k )

        else:
          _context = context

        val[k] = v._eval(
          context = _context,
          logger = logger )

    return type(self)( val, self._loc )

  #-----------------------------------------------------------------------------
  def on_child_key_changed( self, key, new_key ):
    pass




from .base import SchemaPrim
from .pass_prim import PassPrim

from .bool_prim import (
  BoolPrimDeclared,
  BoolPrim )

from .int_prim import (
  IntPrimDeclared,
  IntPrim )

from .float_prim import (
  FloatPrimDeclared,
  FloatPrim )

from .str_prim import (
  StrPrimDeclared,
  StrPrim )

from .seq_prim import (
  SeqPrimDeclared,
  SeqPrim )

from .map_prim import (
  MapPrimDeclared,
  MapPrim )

from .union_prim import (
  UnionPrimDeclared,
  UnionPrim )


seq_declared = SeqPrimDeclared()
map_declared = MapPrimDeclared()

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# bool|float|str|list|dict
any_prim_cases = [
    BoolPrim( default_val = False ),
    FloatPrim( default_val = 0.0 ),
    StrPrim( default_val = "" ),
    seq_declared,
    map_declared ]

AnyPrim = UnionPrim(
  cases = any_prim_cases,
  default_case = 0 )


SeqPrim(
  declared = seq_declared,
  item = AnyPrim,
  default_val = list() )

MapPrim(
  declared = map_declared,
  item = AnyPrim,
  default_val = dict() )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++



# -*- coding: UTF-8 -*-
from collections import OrderedDict as odict
from copy import copy
from abc import ABCMeta
import itertools
import re
import uuid

from abc import ABCMeta

import logging
log = logging.getLogger(__name__)

import ruamel.yaml

from ruamel.yaml.comments import (
  CommentedMap,
  CommentedOrderedMap,
  CommentedKeySeq,
  CommentedSeq )

from ruamel.yaml.scalarstring import LiteralScalarString

from partis.schema_meta.base import (
  Loc,
  SchemaHint,
  SchemaError,
  SchemaParseError,
  SchemaDetectionError,
  is_valued_type,
  is_evaluated,
  is_string,
  is_mapping,
  is_sequence )

from partis.schema.plugin import (
  schema_plugins )

from .utils import (
  as_load,
  as_dump )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def dump_prep( val ):
  """Prepares value to be dumped, converting to ruamel datastructures
  """

  if is_valued_type( val ) or is_evaluated( val ):
    val = val._encode

  if is_mapping( val ):
    _val = CommentedMap()

    for k, v in val.items():
      _val[k] = dump_prep( v )

  elif is_sequence( val ):
    _val = CommentedSeq()

    for v in val:
      _val.append( dump_prep( v ) )

  elif is_string( val ) and len(val.splitlines()) > 1:
    _val = LiteralScalarString(val)

  else:
    _val = val

  return _val

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def loads(
  src,
  schema = None,
  loc = None,
  detect_schema = False ):
  """
  Parse the first YAML source in a stream
  and produce the corresponding Python object.

  Parameters
  ----------
  src : str
  schema : :class:`Schema <partis.schema.struct.SchemaStruct>` | :class:`SchemaPrim <partis.schema.prim.base.SchemaPrim>`
  loc : None | :class:`Loc <partis.schema_meta.base.Loc>`
    Location information of source data (E.G. file, line/column number)
  detect_schema : bool
    If true, attempts to find the schema to use from available plugins by the
    `schema_hash` stored in a top-level ``__schema_hash__`` key.
    If the top level source is not a mapping, or if ``__schema_hash__`` is not
    present, or if the has is not discovered, and ``schema`` (the default)
    is not given, this will raise an error.

  Returns
  -------
  val : object | :class:`Schema <partis.schema.valued.Valued>`
  """

  if loc is None:
    loc = Loc()

  if isinstance( loc, str ):
    loc = Loc( filename = loc )

  try:
    val = ruamel.yaml.round_trip_load( src )

  except BaseException as e:
    raise SchemaParseError(
      f"source failed parsing",
      loc = loc,
      hints = SchemaHint.cast( e ) ) from e

  schema_hash = ''

  if is_mapping( val ) and '__schema_hash__' in val:
    schema_hash = val.pop('__schema_hash__')

  if detect_schema:

    if schema_hash:
      _schemas = schema_plugins.get_by_hash( schema_hash )

      if len(_schemas) > 0:
        schema = _schemas[0]

    if schema is None:
      raise SchemaDetectionError(
        f'Schema could not be detected' )

  if schema is None:
    return val

  loc = loc(val)

  return schema.schema.decode(
    val = val,
    loc = loc )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def dumps(
  val,
  add_hash = False ):
  """Serializes a suitable Python object into a YAML source string

  Parameters
  ----------
  val : object | :class:`Schema <partis.schema.valued.Valued>`

  Returns
  -------
  src : str
  """

  schema = None

  if hasattr( val, '_schema'):
    schema = val._schema

  val = dump_prep( val )

  if add_hash and schema and is_mapping( val ):
    val['__schema_hash__'] = schema.schema_hash

  return ruamel.yaml.round_trip_dump( val,
    # default_style = "|",
    default_flow_style = False,
    allow_unicode = True )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
load = as_load( loads )
dump = as_dump( dumps )



import logging
log = logging.getLogger( __name__ )

from copy import copy
import inspect

from partis.utils import (
  adict,
  odict,
  fmt_obj )

from partis.utils.special import (
  required,
  optional )

from partis.schema_meta.eval import (
  ProviderMeta,
  Provider,
  ProviderSupport,
  EvaluatedContextMeta,
  EvaluatedContext,
  EvaluatedMeta )

from partis.schema_meta.base import (
  SchemaError,
  SchemaEvaluationError,
  SchemaHint,
  Loc,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_evaluated,
  is_valued,
  is_optional,
  is_required )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def _ctx_default( context, context_compat ):
  if context is None:
    # empty 'None' context
    context = EvaluatedContext()

  elif is_mapping( context ):
    # convert regular mappings to a 'None' context with the values
    context = ConstEvaluatedContext( context )

  if not is_sequence( context ):
    context = [ context, ]

  # filter out incompatible contexts
  # NOTE: use of `==` is intended here, instead of `isinstance`, because the
  # equality is compared by the `id` of the respective classes, not a strict
  # inheritance relationship
  context = [ c for c in context if type(c) == context_compat ]

  return context

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class ConstEvaluatedContext( EvaluatedContext ):
  """Constant namespace for all evaluations
  """
  #-----------------------------------------------------------------------------
  def __init__( self, locals = None ):
    super().__init__()

    self._p_locals = dict(locals) if locals else dict()

  #-----------------------------------------------------------------------------
  def locals( self, schema ):

    return self._p_locals

  #-----------------------------------------------------------------------------
  def __call__( self,
    schema,
    parent,
    key ):

    return self

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class Evaluated( metaclass = EvaluatedMeta ):
  """Base class for evaluated sources

  Parameters
  ----------
  schema : :class:`Schema <partis.schema.Schema>`
  src : object
    Source data to be evaluated.
  loc : :class:`Loc <partis.schema_meta.base.Loc>`
    Location information of source data (E.G. file, line/column number)

  Note
  ----
  If an expression returns `None`, the default value specified by the schema
  will be used as the result of the expression.
  This could cause an error when the expression evaluates to None, and the default
  value of the schema is also an expression.
  This could also cause an error if the default value, if it is an expression,
  evaluated to None, since this would then result in the same un-evaluated expression.

  """

  #-----------------------------------------------------------------------------
  def __init__( self,
    schema,
    src,
    loc = None ):

    if not is_schema( schema ):
      raise SchemaError(
        f"`schema` must be instance of `Schema` or `SchemaPrim`: {type(schema)}")

    if loc is None:
      loc = Loc()

    if not isinstance( loc, Loc ):
      raise SchemaError(
        f"`loc` must be instance of `Loc`: {type(loc)}")

    self._p_schema = schema
    self._p_src = src
    self._p_loc = loc

  #-----------------------------------------------------------------------------
  def __eq__( self, other ):
    return (
      type(self) == type(other)
      and self._schema == other._schema
      and self._src == other._src )

  #-----------------------------------------------------------------------------
  def __str__( self ):
    return f"{type(self).__name__}({self._src})"

  #-----------------------------------------------------------------------------
  def __repr__( self ):
    return str(self)

  #-----------------------------------------------------------------------------
  @property
  def _provider( self ):
    return type(self).provider

  #-----------------------------------------------------------------------------
  @property
  def _context( self ):
    return type(self).context

  #-----------------------------------------------------------------------------
  @property
  def _schema( self ):
    return self._p_schema

  #-----------------------------------------------------------------------------
  @property
  def _src( self ):
    return self._p_src

  #-----------------------------------------------------------------------------
  @property
  def _loc( self ):
    return self._p_loc

  #-----------------------------------------------------------------------------
  @property
  def _encode( self ):
    return type(self).escaped( *(type(self).check( self._src )) )

  #-----------------------------------------------------------------------------
  def _lint( self,
    context = None,
    logger = None ):
    """Lints source

    Parameters
    ----------
    context : None | :class:`EvaluatedContext` | list[ :class:`EvaluatedContext` ]
      Sets the context for which this expression is to be evaluated.
      The expression is evaluated only when the specified context is equivalent
      to the one for this evaluated class.
    logger : logging.Logger
      A logger object to direct expression logs/prints instead of stdout

    Returns
    -------
    : list[SchemaHint]

    """

    context = _ctx_default( context, self._context )

    if len(context) == 0:
      # only evaluate if specified context does not exclude this evaluation
      return list()

    context = next(iter(context))

    if self._provider is None:
      # TODO: why is this not done in the initializer?
      raise SchemaError(f"May not perform evaluation without a `provider`")

    try:
      hints = self._provider.lint(
        schema = self._schema,
        src = self._src,
        locals = adict( context.locals(
          schema = self._schema ) ),
        module = context.module,
        logger = logger )

    except BaseException as e:

      raise SchemaEvaluationError(
        f"Linting failed",
        loc = self._loc,
        hints = [
          f"source: {self._src}",
          f"context: {context}",
          SchemaHint.cast( e ) ] ) from e

    return hints

  #-----------------------------------------------------------------------------
  def _eval( self,
    context = None,
    logger = None ):
    """Evaluates source

    Parameters
    ----------
    context : None | :class:`EvaluatedContext` | list[ :class:`EvaluatedContext` ]
      Sets the context for which this expression is to be evaluated.
      The expression is evaluated only when the specified context is equivalent
      to the one for this evaluated class.
    logger : logging.Logger
      A logger object to direct expression logs/prints instead of stdout

    Returns
    -------
    : object

    """

    context = _ctx_default( context, self._context )

    if len(context) == 0:
      # only evaluate if specified context does not exclude this evaluation
      return self

    context = next(iter(context))

    if self._provider is None:
      # TODO: why is this not done in the initializer?
      raise SchemaError(f"May not perform evaluation without a `provider`")

    try:
      _val = self._provider.eval(
        schema = self._schema,
        src = self._src,
        locals = adict( context.locals(
          schema = self._schema ) ),
        module = context.module,
        logger = logger )

    except BaseException as e:

      raise SchemaEvaluationError(
        f"Evaluation failed",
        loc = self._loc,
        hints = [
          f"source: {self._src}",
          f"context: {context}",
          SchemaHint.cast( e ) ] ) from e

    try:
      val = self._schema.decode(
        val = _val,
        loc = self._loc )

    except BaseException as e:

      raise SchemaEvaluationError(
        f"Evaluation result not decoded by schema",
        loc = self._loc,
        hints = [
          f"source: {self._src}",
          f"context: {context}",
          f"result: {fmt_obj(_val)}",
          SchemaHint.cast( e ) ] ) from e

    if not is_valued(val):
      if (
        is_evaluated( val._src )
        and _val is None
        and val._src._src == self._src ):

        # NOTE: this means the default value is itself an expression, and the source
        # code is the same, so evaluating the default value will currently cause
        # infinite recursion.

        if is_required( self._schema.default_eval ):
          # No default eval given to use in this case

          raise SchemaEvaluationError(
            f"Evaluation resulted in another evaluation from default values",
            loc = self._loc,
            hints = [
              f"source: {self._src}",
              f"context: {context}",
              f"result: {fmt_obj(_val)}",
              f"decoded: {fmt_obj(val)}" ] )

        val = self._schema.default_eval

        # the machinery should have ensured default_eval is valued
        assert is_valued(val)

      else:
        # NOTE: It's not clear the constraints that should be placed on recursively
        # evaluating expressions. For instance, an expression that evaluates to
        # a path that then evaluates to a relative/absolute path would be ok, but
        # maybe not to another general expression.
        val = val._eval(
          context = context,
          logger = logger )

    return val

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class NotEvaluated( Evaluated ):
  pass



import pytest
import numpy as np
import array

import itertools

from partis.schema import (
  required,
  optional,
  BoolPrim,
  IntPrim,
  FloatPrim,
  StrPrim,
  SeqPrim,
  MapPrim,
  UnionPrim,
  StructValued,
  PJCEvaluated,
  SchemaError,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_declared,
  is_schema_struct_valued,
  is_valued_type,
  is_valued,
  is_evaluated,
  is_similar_value_type,
  any_schema,
  schema_declared,
  SchemaDefinitionError,
  SchemaDeclaredError,
  PJCEvaluated )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_declared():

  d = schema_declared( tag = 'my_type1', tag_key = None )

  assert( d.tag_key == 'type' )
  assert( d.tag == 'my_type1' )

  with pytest.raises( SchemaDeclaredError ):
    d = schema_declared( tag = None )

  with pytest.raises( SchemaDeclaredError ):
    d = schema_declared( tag = '1' )

  with pytest.raises( SchemaDeclaredError ):
    d = schema_declared( tag = 'mytag', tag_key = '1' )

  with pytest.raises( SchemaDeclaredError ):
    d = schema_declared( tag = 'mytag', tag_key = ['a', None] )

  with pytest.raises( SchemaDeclaredError ):
    d = schema_declared( tag = 'mytag', tag_key = ['a', '1'] )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_declared_defined():

  d1 = schema_declared( tag = 'my_type1', tag_key = 'mytagkey' )
  d2 = schema_declared( tag = 'my_type2', tag_key = None )

  assert( not d1.schema_defined )
  assert( not d2.schema_defined )

  with pytest.raises( SchemaDeclaredError ):
    # must define using a schema class
    d1.schema_declared( None )

  with pytest.raises( SchemaDeclaredError ):
    # cannot be another declared
    d1.schema_declared( d2 )

  class MyType1( StructValued,
    declared = d1,
    struct = [ ] ):
    pass

  class MyType2( StructValued,
    tag = 'my_type2',
    tag_key = 'type',
    struct = [ ] ):
    pass

  class MyType3( StructValued,
    tag = 'my_type2',
    # mismatch key
    tag_key = 'type3',
    struct = [ ] ):
    pass

  class MyType4( StructValued,
    # mismatch tag
    tag = 'my_type3',
    tag_key = 'type',
    struct = [ ] ):
    pass


  with pytest.raises( SchemaDeclaredError ):
    # already defined
    d1.schema_declared( MyType2 )

  with pytest.raises( SchemaDeclaredError ):
    # tag/key mismatch
    d2.schema_declared( MyType3 )

  with pytest.raises( SchemaDeclaredError ):
    # tag/key mismatch
    d2.schema_declared( MyType4 )

  # should be ok
  d2.schema_declared( MyType2 )

  # should now be defined
  assert( d1.schema_defined )
  assert( d1.schema is MyType1.schema )
  assert( MyType1.tag == 'my_type1' )
  assert( MyType1.tag_key == 'mytagkey' )


  assert( d2.schema_defined )
  assert( d2.schema is MyType2.schema )



import pytest
from pytest import raises

from pprint import pprint

import itertools

from partis.schema import (
  Loc,
  required,
  optional,
  BoolPrim,
  IntPrim,
  FloatPrim,
  StrPrim,
  SeqPrim,
  MapPrim,
  UnionPrim,
  SchemaStruct,
  PJCEvaluated,
  NotEvaluated,
  SchemaError,
  SchemaDefinitionError,
  SchemaValidationError,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  any_schema, )

from partis.schema.serialize.yaml import (
  loads,
  dumps )

from test_prim import (
  is_defined,
  get_val,
  should_pass,
  max_tests,
  step_tests )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# use this to only run sub-set of all possible tests
# max_tests = -1
# step_tests = 1

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def prim_list_cases( ):
  vals = [
    None,
    "$expr:py self",
    "$func:py return self",
    [ 1, 2, 3 ],
    [ 'a', 'b', 'c' ] ]

  items = [
    ( IntPrim(), is_numeric ),
    ( FloatPrim(), is_numeric ),
    ( StrPrim(), is_string ) ]

  default_vals = [
    optional,
    required,
    [ 4, 5, 6 ],
    [ 'x', 'y', 'z' ] ]

  restricteds = [
    None ]

  evaluateds = [
    None,
    PJCEvaluated ]

  test_cases = itertools.product(
    vals,
    items,
    default_vals,
    restricteds,
    evaluateds )

  if max_tests >= 0:
    test_cases = itertools.islice( test_cases, 0, max_tests * step_tests, step_tests )

  return test_cases

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
@pytest.fixture( params = prim_list_cases() )
def prim_list_case( request ):
  return request.param

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_prim_list( prim_list_case ):

  val, ( item, is_item ), default_val, restricted, evaluated = prim_list_case

  if should_pass(
    val = val,
    is_type = is_sequence,
    is_item = is_item,
    default_val = default_val,
    restricted = restricted,
    evaluated = evaluated ):

    p = SeqPrim(
      item = item,
      default_val = default_val,
      restricted = restricted,
      evaluated = evaluated,
      doc = "My list" )

    assert p.default_val == default_val
    assert p.restricted == restricted
    assert p.doc.startswith("My list")
    assert p.evaluated == NotEvaluated if evaluated is None else evaluated

    _val = p.decode( val )

    if evaluated is not None and evaluated.check(val):
      if is_defined(default_val):
        # only check evaluated when default value is defined
        _val = _val._eval( context = { 'self' : default_val } )
        assert _val == p.default_val

    else:
      assert _val == get_val( val, default_val )

  else:
    with pytest.raises(SchemaError):

      p = SeqPrim(
        item = item,
        default_val = default_val,
        restricted = restricted,
        evaluated = evaluated,
        doc = "My list" )

      _val = p.decode( val )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_prim_seq_default():
  loc = Loc()

  p = SeqPrim(
    item = BoolPrim() )

  val = [ True, False ]
  v = p.decode(val)
  w = p(val)

  assert v == val
  assert w == val
  assert w == v
  assert v._encode == val
  assert w._encode == val
  assert p.encode(v) == val
  assert p.encode(w) == val


  with raises(SchemaDefinitionError):
    # not a schema
    SeqPrim(
      item = int )



# -*- coding: UTF-8 -*-

from .base import (
  SchemaPrimDeclared,
  SchemaPrim )

from .pass_prim import PassPrim

from .bool_prim import (
  BoolPrimDeclared,
  BoolPrim )

from .int_prim import (
  IntPrimDeclared,
  IntPrim )

from .float_prim import (
  FloatPrimDeclared,
  FloatPrim )

from .str_prim import (
  StrPrimDeclared,
  StrPrim )

from .seq_prim import (
  SeqPrimDeclared,
  SeqPrim )

from .map_prim import (
  MapPrimDeclared,
  MapPrim )

from .union_prim import (
  UnionPrimDeclared,
  UnionPrim )




import logging
log = logging.getLogger(__name__)

from partis.schema_meta.base import (
  SchemaError,
  SchemaHint,
  assert_valid_name )

from partis.schema_meta.struct import (
  SchemaStructDeclared )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaDeclared:
  """A schema-declared referencing schemas not yet defined
  """

  #-----------------------------------------------------------------------------
  def __init__( self,
    val,
    loc = None ):

    raise SchemaError(
      f"This is a schema declared for `{self.__class__.tag}`, and may not be instantiated.")

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def schema_declared(
  tag,
  tag_key = None ):

  class NewSchemaDeclared( SchemaDeclared,
    tag = tag,
    tag_key = tag_key,
    metaclass = SchemaStructDeclared ):
    pass

  return NewSchemaDeclared



from partis.utils.sphinx import basic_main

import os
import os.path as osp

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
if __name__ == "__main__":

  conf_dir = osp.abspath( osp.dirname(__file__) )
  root_dir = osp.abspath( osp.join( conf_dir, os.pardir ) )

  basic_main(
    package = 'partis-schema',
    conf_dir = conf_dir,
    src_dir = conf_dir,
    root_dir = root_dir )



import pytest

from pprint import pprint

import itertools

from partis.schema import (
  required,
  optional,
  BoolPrim,
  IntPrim,
  FloatPrim,
  StrPrim,
  SeqPrim,
  MapPrim,
  UnionPrim,
  SchemaStruct,
  PJCEvaluated,
  NotEvaluated,
  SchemaError,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  any_schema, )

from partis.schema.serialize.yaml import (
  loads,
  dumps )

from test_prim import (
  is_defined,
  get_val,
  should_pass,
  max_tests,
  step_tests )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# use this to only run sub-set of all possible tests
# max_tests = -1
# step_tests = 1

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def prim_int_cases( ):
  default_vals = [
    optional,
    required,
    0,
    5,
    -10,
    11 ]

  mins = [
    None,
    -10,
    0,
    10 ]

  maxs = [
    None,
    -10,
    0,
    10 ]

  restricteds = [
    None,
    [0],
    [],
    [0, 10, 11] ]

  evaluateds = [
    None,
    PJCEvaluated ]

  vals = [
    None,
    "$expr:py self",
    "$func:py return self",
    -11,
    -10,
    0,
    10,
    11 ]

  test_cases = itertools.product(
    vals,
    default_vals,
    mins,
    maxs,
    restricteds,
    evaluateds )

  if max_tests >= 0:
    test_cases = itertools.islice( test_cases, 0, max_tests * step_tests, step_tests )

  return test_cases



#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
@pytest.fixture( params = prim_int_cases() )
def prim_int_case( request ):
  return request.param


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_prim_int( prim_int_case ):

  val, default_val, min, max, restricted, evaluated = prim_int_case

  if should_pass(
    val = val,
    is_type = is_numeric,
    default_val = default_val,
    restricted = restricted,
    evaluated = evaluated,
    min = min,
    max = max ):

    p = IntPrim(
      default_val = default_val,
      min = min,
      max = max,
      restricted = restricted,
      evaluated = evaluated,
      doc = "My integer" )

    assert p.default_val == default_val
    assert p.min == min
    assert p.max == max
    assert p.restricted == restricted
    assert p.doc.startswith("My integer")
    assert p.evaluated == NotEvaluated if evaluated is None else evaluated

    _val = p.decode( val )

    if evaluated is not None and evaluated.check(val):
      if is_defined(default_val):
        _val = _val._eval( context = { 'self' : default_val } )
        assert _val == p.default_val

    else:
      assert _val == get_val( val, default_val )


  else:
    with pytest.raises(SchemaError):

      p = IntPrim(
        default_val = default_val,
        min = min,
        max = max,
        restricted = restricted,
        evaluated = evaluated,
        doc = "My integer" )

      _val = p.decode( val )



import types
import importlib
import importlib.abc
import inspect
import sys

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaLoader( importlib.abc.Loader ):
  """Loader for a schema module
  """
  #-----------------------------------------------------------------------------
  def __init__( self, schema_module ):

    self._p_schema_module = schema_module

  #-----------------------------------------------------------------------------
  def create_module( self, spec ):

    return self._p_schema_module

  #-----------------------------------------------------------------------------
  def exec_module( self, module ):

    return module


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaFinder( importlib.abc.MetaPathFinder ):
  """Finder for schema modules
  """
  #-----------------------------------------------------------------------------
  def __init__( self ):
    self._p_schema_specs = dict()

  #-----------------------------------------------------------------------------
  @property
  def schema_specs( self ):
    return self._p_schema_specs

  #-----------------------------------------------------------------------------
  def register( self, schema_module, origin = None ):

    if not isinstance( schema_module, SchemaModule ):
      raise ValueError(f"`schema_module` must be instance of SchemaModule: {type(schema_module)}")

    fullname = schema_module.__name__

    if fullname in self.schema_specs:
      raise ValueError(f"Schema module already registered: {fullname}")

    spec = importlib.machinery.ModuleSpec(
      fullname,
      loader = SchemaLoader(
        schema_module = schema_module ),
      origin = origin,
      loader_state = None,
      is_package = False  )

    self.schema_specs[ fullname ] = spec

    return spec

  #-----------------------------------------------------------------------------
  def find_spec( self, fullname, path, target = None ):

    if fullname not in self._p_schema_specs:
      return None

    return self._p_schema_specs[ fullname ]

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
schema_finder = SchemaFinder()

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
sys.meta_path.append( schema_finder )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaModule ( types.ModuleType ):
  """Base class for schema modules
  """

  #-----------------------------------------------------------------------------
  def __init__( self, fullname = None, doc = None, origin = None ):

    if fullname is None:
      idx = 0

      fullname = f"{type(self).__name__}_{idx}"

      while fullname in schema_finder.schema_specs:
        idx += 1
        fullname = f"{type(self).__name__}_{idx}"

    super().__init__( fullname, doc = doc )

    spec = schema_finder.register( self, origin = origin )
    self.__loader__ = spec.loader

    if origin is not None:
      self.__file__ = origin



"""Usage of `setup.py` is deprecated, and is supplied only for legacy installation.
"""
import sys
import os
import os.path as osp
import importlib
import logging
import argparse
import subprocess
import tempfile
from argparse import RawTextHelpFormatter
logger = logging.getLogger(__name__)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def egg_info( args ):

  logger.warning(
    "running legacy 'setup.py egg_info'" )

  dir = osp.join( args.egg_base, EGG_INFO_NAME )

  if not osp.exists( dir ):
    os.mkdir( dir )

  with open( osp.join( dir, 'PKG-INFO' ), 'wb' ) as fp:
    fp.write( PKG_INFO )

  with open( osp.join( dir, 'setup_requires.txt' ), 'wb' ) as fp:
    fp.write( b'' )

  with open( osp.join( dir, 'requires.txt' ), 'wb' ) as fp:
    fp.write( REQUIRES )

  with open( osp.join( dir, 'SOURCES.txt' ), 'wb' ) as fp:
    fp.write( SOURCES )

  with open( osp.join( dir, 'top_level.txt' ), 'wb' ) as fp:
    fp.write( TOP_LEVEL )

  with open( osp.join( dir, 'entry_points.txt' ), 'wb' ) as fp:
    fp.write( ENTRY_POINTS )

  with open( osp.join( dir, 'dependency_links.txt' ), 'wb' ) as fp:
    fp.write( b'' )

  with open( osp.join( dir, 'not-zip-safe' ), 'wb' ) as fp:
    fp.write( b'' )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def bdist_wheel( args ):

  logger.warning(
    "running legacy 'setup.py bdist_wheel'" )

  sys.path = backend_path + sys.path

  backend = importlib.import_module( build_backend )

  backend.build_wheel(
    wheel_directory = args.dist_dir or args.bdist_dir or '.' )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def install( args ):

  logger.warning(
    "running legacy 'setup.py install'" )

  reqs = [ f"{r}" for r in build_requires ]

  subprocess.check_call([
    sys.executable,
    '-m',
    'pip',
    'install',
    *reqs ] )

  sys.path = backend_path + sys.path

  backend = importlib.import_module( build_backend )

  with tempfile.TemporaryDirectory() as tmpdir:
    wheel_name = backend.build_wheel(
      wheel_directory = tmpdir )

    subprocess.check_call([
      sys.executable,
      '-m',
      'pip',
      'install',
      osp.join(tmpdir, wheel_name) ])

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def dummy( args ):
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def main():

  logging.basicConfig(
    level = logging.INFO,
    format = "{name}:{levelname}: {message}",
    style = "{" )


  logger.warning(
    "'setup.py' is deprecated, limited support for legacy installs. Upgrade pip." )

  parser = argparse.ArgumentParser(
    description = __doc__,
    formatter_class = RawTextHelpFormatter )

  subparsers = parser.add_subparsers()

  #.............................................................................
  egg_info_parser = subparsers.add_parser( 'egg_info' )

  egg_info_parser.set_defaults( func = egg_info )

  egg_info_parser.add_argument( "-e", "--egg-base",
    type = str,
    default = '.' )

  #.............................................................................
  bdist_wheel_parser = subparsers.add_parser( 'bdist_wheel' )

  bdist_wheel_parser.set_defaults( func = bdist_wheel )

  bdist_wheel_parser.add_argument( "-b", "--bdist-dir",
    type = str,
    default = '' )

  bdist_wheel_parser.add_argument( "-d", "--dist-dir",
    type = str,
    default = '' )

  bdist_wheel_parser.add_argument( "--python-tag",
    type = str,
    default = None )

  bdist_wheel_parser.add_argument( "--plat-name",
    type = str,
    default = None )

  bdist_wheel_parser.add_argument( "--py-limited-api",
    type = str,
    default = None )

  bdist_wheel_parser.add_argument( "--build-number",
    type = str,
    default = None )

  #.............................................................................
  install_parser = subparsers.add_parser( 'install' )

  install_parser.set_defaults( func = install )

  install_parser.add_argument( "--record",
    type = str,
    default = None )

  install_parser.add_argument( "--install-headers",
    type = str,
    default = None )

  install_parser.add_argument( "--compile",
    action='store_true' )

  install_parser.add_argument( "--single-version-externally-managed",
    action='store_true' )

  #.............................................................................
  clean_parser = subparsers.add_parser( 'clean' )

  clean_parser.set_defaults( func = dummy )

  clean_parser.add_argument( "-a", "--all",
    action='store_true' )

  args = parser.parse_args( )

  args.func( args )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# NOTE: these are templated literal values substituded by the backend when
# building the source distribution

build_backend = 'partis.pyproj.backend'
backend_path = []
build_requires = ['partis-pyproj==0.0.1', 'wheel']

EGG_INFO_NAME = 'partis-schema.egg-info'

PKG_INFO = b'Metadata-Version: 2.1\nName: partis-schema\nVersion: 0.0.1\nRequires-Python: >=3.6.2\nMaintainer-email: "Nanohmics Inc." <software.support@nanohmics.com>\nSummary: Base classes for defining JSON-compatible schemas\nLicense-File: LICENSE.txt\nClassifier: License :: OSI Approved :: BSD License\nClassifier: Topic :: Software Development :: Libraries\nClassifier: Development Status :: 4 - Beta\nClassifier: Operating System :: Microsoft :: Windows\nClassifier: Intended Audience :: Developers\nClassifier: Programming Language :: Python\nClassifier: Operating System :: POSIX :: Linux\nClassifier: Programming Language :: Python :: 3\nRequires-Dist: partis-pyproj==0.0.1\nRequires-Dist: tomli>=1.2.3\nRequires-Dist: ruamel.yaml==0.16.5\nRequires-Dist: pygments==2.9.0\nRequires-Dist: cheetah3<4.0.0,>=3.2.4\nRequires-Dist: partis-utils[lint]==0.0.1\nRequires-Dist: overrides<4.0.0,>=3.1.0\nRequires-Dist: wheel\nDescription-Content-Type: text/x-rst\n\nThe ``partis.schema`` package contains base classes for defining JSON-compatible schemas.'

REQUIRES = b'partis-pyproj==0.0.1\ntomli>=1.2.3\nruamel.yaml==0.16.5\npygments==2.9.0\ncheetah3<4.0.0,>=3.2.4\npartis-utils[lint]==0.0.1\noverrides<4.0.0,>=3.1.0\nwheel'

SOURCES = b'partis_schema-0.0.1/src/schema/prim/seq_prim.py\npartis_schema-0.0.1/src/schema/prim/__init__.py\npartis_schema-0.0.1/src/schema/prim/pass_prim.py\npartis_schema-0.0.1/src/schema/prim/map_prim.py\npartis_schema-0.0.1/src/schema/prim/base.py\npartis_schema-0.0.1/src/schema/prim/bool_prim.py\npartis_schema-0.0.1/src/schema/prim/any_prim.py\npartis_schema-0.0.1/src/schema/prim/int_prim.py\npartis_schema-0.0.1/src/schema/prim/str_prim.py\npartis_schema-0.0.1/src/schema/prim/union_prim.py\npartis_schema-0.0.1/src/schema/prim/float_prim.py\npartis_schema-0.0.1/src/schema/__init__.py\npartis_schema-0.0.1/src/schema/declared.py\npartis_schema-0.0.1/src/schema/eval/__init__.py\npartis_schema-0.0.1/src/schema/eval/eval.py\npartis_schema-0.0.1/src/schema/eval/eval_cheetah.py\npartis_schema-0.0.1/src/schema/eval/eval_py.py\npartis_schema-0.0.1/src/schema/eval/eval_pjc.py\npartis_schema-0.0.1/src/schema/serialize/__init__.py\npartis_schema-0.0.1/src/schema/serialize/config.py\npartis_schema-0.0.1/src/schema/serialize/yaml.py\npartis_schema-0.0.1/src/schema/serialize/json.py\npartis_schema-0.0.1/src/schema/serialize/utils.py\npartis_schema-0.0.1/src/schema/module.py\npartis_schema-0.0.1/src/schema/hint.py\npartis_schema-0.0.1/src/schema/plugin.py\npartis_schema-0.0.1/src/schema/valued/struct_valued.py\npartis_schema-0.0.1/src/schema/valued/__init__.py\npartis_schema-0.0.1/src/schema/valued/str_valued.py\npartis_schema-0.0.1/src/schema/valued/int_valued.py\npartis_schema-0.0.1/src/schema/valued/map_valued.py\npartis_schema-0.0.1/src/schema/valued/valued.py\npartis_schema-0.0.1/src/schema/valued/float_valued.py\npartis_schema-0.0.1/src/schema/valued/bool_valued.py\npartis_schema-0.0.1/src/schema/valued/seq_valued.py\npartis_schema-0.0.1/src/schema_meta/__init__.py\npartis_schema-0.0.1/src/schema_meta/struct.py\npartis_schema-0.0.1/src/schema_meta/eval.py\npartis_schema-0.0.1/src/schema_meta/base.py\npartis_schema-0.0.1/src/schema_meta/property.py\npartis_schema-0.0.1/src/schema_meta/schema.py\npartis_schema-0.0.1/src/schema_meta/valued.py\npartis_schema-0.0.1/src/schema_meta/prim.py\npartis_schema-0.0.1/doc/conf.py\npartis_schema-0.0.1/doc/__init__.py\npartis_schema-0.0.1/doc/index.rst\npartis_schema-0.0.1/doc/src/partis.schema.prim.union_prim.rst\npartis_schema-0.0.1/doc/src/partis.schema_meta.schema.rst\npartis_schema-0.0.1/doc/src/partis.schema_meta.struct.rst\npartis_schema-0.0.1/doc/src/partis.schema.prim.float_prim.rst\npartis_schema-0.0.1/doc/src/partis.schema_meta.rst\npartis_schema-0.0.1/doc/src/partis.schema_meta.prim.rst\npartis_schema-0.0.1/doc/src/partis.schema.prim.map_prim.rst\npartis_schema-0.0.1/doc/src/partis.schema.eval.rst\npartis_schema-0.0.1/doc/src/partis.schema.valued.rst\npartis_schema-0.0.1/doc/src/partis.schema_meta.property.rst\npartis_schema-0.0.1/doc/src/partis.schema.prim.str_prim.rst\npartis_schema-0.0.1/doc/src/partis.schema.prim.seq_prim.rst\npartis_schema-0.0.1/doc/src/partis.schema.plugin.rst\npartis_schema-0.0.1/doc/src/partis.schema.module.rst\npartis_schema-0.0.1/doc/src/partis.schema.prim.bool_prim.rst\npartis_schema-0.0.1/doc/src/partis.schema.serialize.rst\npartis_schema-0.0.1/doc/src/partis.schema.prim.base.rst\npartis_schema-0.0.1/doc/src/partis.schema.rst\npartis_schema-0.0.1/doc/src/partis.schema_meta.eval.rst\npartis_schema-0.0.1/doc/src/partis.schema_meta.base.rst\npartis_schema-0.0.1/doc/src/partis.schema.declared.rst\npartis_schema-0.0.1/doc/src/partis.schema.prim.rst\npartis_schema-0.0.1/doc/src/partis.schema_meta.valued.rst\npartis_schema-0.0.1/doc/src/partis.schema.prim.int_prim.rst\npartis_schema-0.0.1/doc/src/partis.schema.prim.pass_prim.rst\npartis_schema-0.0.1/doc/__main__.py\npartis_schema-0.0.1/test/300_schema/test_prim_str.py\npartis_schema-0.0.1/test/300_schema/test_prim_float.py\npartis_schema-0.0.1/test/300_schema/__init__.py\npartis_schema-0.0.1/test/300_schema/test_prim_map.py\npartis_schema-0.0.1/test/300_schema/test_valued.py\npartis_schema-0.0.1/test/300_schema/test_struct.py\npartis_schema-0.0.1/test/300_schema/test_prim_union.py\npartis_schema-0.0.1/test/300_schema/test_prim_bool.py\npartis_schema-0.0.1/test/300_schema/test_prim.py\npartis_schema-0.0.1/test/300_schema/test_declared.py\npartis_schema-0.0.1/test/300_schema/test_hint.py\npartis_schema-0.0.1/test/300_schema/test_prim_int.py\npartis_schema-0.0.1/test/300_schema/test_eval_pjc.py\npartis_schema-0.0.1/test/300_schema/test_prim_seq.py\npartis_schema-0.0.1/test/300_schema/.coverage\npartis_schema-0.0.1/test/200_schema_meta/test_203_struct.py\npartis_schema-0.0.1/test/200_schema_meta/__init__.py\npartis_schema-0.0.1/test/200_schema_meta/test_201_base.py\npartis_schema-0.0.1/test/200_schema_meta/test_202_declared.py\npartis_schema-0.0.1/pyproject.toml\npartis_schema-0.0.1/LICENSE.txt\npartis_schema-0.0.1/README.rst'

TOP_LEVEL = b''

ENTRY_POINTS = b''

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

if __name__ == "__main__":
  exit( main() )



# -*- coding: UTF-8 -*-

import logging
log = logging.getLogger(__name__)

from copy import copy
import inspect
import re

from partis.utils import (
  odict,
  fmt_base_or_type,
  fmt_iterable_or )


from partis.schema_meta.base import (
  SchemaError,
  SchemaNameError,
  SchemaDeclaredError,
  SchemaDefinitionError,
  SchemaValidationError,
  SchemaHint,
  Loc,
  assert_valid_name,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_evaluated,
  is_valued,
  is_optional,
  is_required )

from partis.schema.valued import (
  BoolValued )

from . import (
  SchemaPrimDeclared,
  SchemaPrim )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class BoolPrimDeclared( SchemaPrimDeclared ):
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class BoolPrim( SchemaPrim ):
  """Primitive for boolean values

  Parameters
  ----------
  **kwargs : arguments passed to :class:`SchemaPrim <partis.schema.prim.SchemaPrim>`
  """

  #-----------------------------------------------------------------------------
  def __new__( mcls,
    name,
    bases,
    namespace,
    **kwargs ):


    return super().__new__( mcls,
      name = name,
      bases = bases,
      namespace = namespace,
      valued_type = BoolValued,
      **kwargs )

  #-----------------------------------------------------------------------------
  def validate( cls,
    val,
    loc = None ):

    if not is_bool( val ):

      raise SchemaValidationError(
        f"Must be a `bool`: {type(val).__name__}",
        loc = loc )

    val = bool(val)

    return super().validate(
      val = val,
      loc = loc )



import pytest

from pprint import pprint

import itertools

from partis.schema import (
  required,
  optional,
  BoolPrim,
  IntPrim,
  FloatPrim,
  StrPrim,
  SeqPrim,
  MapPrim,
  UnionPrim,
  StructValued,
  PJCEvaluated,
  SchemaError,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  any_schema,
  SchemaDefinitionError )

from partis.schema.serialize.yaml import (
  loads,
  dumps )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_schema():

  # test using metaclass arguments
  class MyType1( StructValued,
    tag = 'my_type',
    struct = [
      ( 'a', BoolPrim() ),
      ( 'b', IntPrim() ),
      ( 'c', FloatPrim() ),
      ( 'd', StrPrim() ),
      ( 'e', SeqPrim(
        item = StrPrim() ) ),
      ( 'f', MapPrim(
        item = StrPrim() ) )  ] ):
    pass

  # test using class namespace arguments
  class MyType2( StructValued ):
    schema = dict(
      tag = 'my_type' )

    a = BoolPrim()
    b = IntPrim()
    c = FloatPrim()
    d = StrPrim()
    e =  SeqPrim(
      item = StrPrim() )

    f = MapPrim(
      item = StrPrim() )

  doc_yaml = """{ type: my_type, a: True, b: 1, c: 1.0, d: "one", e: ["x", "y", "z"], f: { q: q, r: r} }"""

  for cls in [ MyType1, MyType2 ]:
    obj = loads( doc_yaml, cls )

    assert isinstance( obj, cls )
    assert obj._schema is cls.schema
    assert obj._schema.tag == 'my_type'
    assert obj.a == True
    assert obj.b == 1
    assert obj.c == 1.0
    assert obj.d == "one"
    assert obj.e == ["x", "y", "z"]
    assert obj.f == { "q": "q", "r": "r"}

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_struct_proxy():

  # test using metaclass arguments
  class MyType1( StructValued,
    tag = 'my_type',
    struct_proxy = 'a',
    struct = [
      ( 'a', BoolPrim() ),
      ( 'b', IntPrim( default_val = 1 ) ) ] ):
    pass

  x = MyType1( True )

  assert x.a == True
  assert x.b == 1

  with pytest.raises( SchemaDefinitionError ):
    # should raise error because struct_proxy conflicts with union
    u = UnionPrim(
      cases = [
        MyType1,
        BoolPrim() ] )



# -*- coding: UTF-8 -*-
import os
import sys
import inspect
import weakref
from inspect import getframeinfo, stack
import traceback
import linecache
from copy import copy, deepcopy
import importlib

import logging
log = logging.getLogger(__name__)

from partis.utils import (
  indent_lines,
  line_segment )

# linting python source
from partis.utils.lint import (
  lint_python )

from partis.schema_meta.base import (
  SchemaHint,
  Loc,
  SchemaEvaluationError )

from .eval import (
  ProviderSupport,
  Provider,
  Evaluated )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def extract_exec_stack( e, filename, funcname ):
  # extract traceback information

  _args = ", ".join([ str(arg) for arg in e.args ])

  hint = SchemaHint(
    f"{e.__class__.__name__}: { _args }",
    level = 'error' )

  prev_hint = hint

  _stack = list( traceback.walk_tb( e.__traceback__ ) )

  line_offset = 0

  # first determin the offset of the function body for the user's code
  for frame, lineno in _stack:
    code = frame.f_code

    if code.co_filename == filename:
      if code.co_name == funcname:
        line_offset = code.co_firstlineno
        break

  # build trace starting at deepest point
  for frame, lineno in _stack[::-1]:

    code = frame.f_code

    if code.co_filename == filename:
      # remove the offset line-number to the function

      if code.co_name == funcname:
        # stop as soon as traceback reaches main code level
        return SchemaHint(
          f"In compiled source:",
          loc = f"line {lineno-line_offset}, `{linecache.getline( code.co_filename, lineno ).strip()}`",
          level = 'error',
          hints = prev_hint )

    else:

      prev_hint = SchemaHint(
        f"In function: `{code.co_name}`",
        loc = f"File \"{code.co_filename}\", line {lineno}",
        level = 'error',
        hints = SchemaHint(
          f"{os.path.basename(code.co_filename)}:{lineno} `{linecache.getline( code.co_filename, lineno ).strip()}`",
          level = 'error',
          hints = prev_hint ) )

  return prev_hint

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class EvalFunc:
  #-----------------------------------------------------------------------------
  def __init__( self, func ):
    if not callable(func):
      raise ValueError(f"`func` must be callable: {type(func).__name__}")

    self._func = func

  #-----------------------------------------------------------------------------
  def __call__( self, **kwargs ):
    return self._func( **kwargs )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class PyProvider( Provider ):
  """Evaluates python expressions
  """

  TAG_EXPR_PY = "$expr:py"
  TAG_FUNC_PY = "$func:py"

  #-----------------------------------------------------------------------------
  def __init__( self ):
    super().__init__()

    self._supported = dict(
      expr_py = ProviderSupport(
        name = "Python expression",
        lexer = 'python',
        doc = f"escaped string: `{self.TAG_EXPR_PY} ...`" ),
      func_py = ProviderSupport(
        name = "Python function",
        lexer = 'python',
        doc = f"escaped string: `{self.TAG_FUNC_PY} ...`" ),
      call_py = ProviderSupport(
        name = "Python callable",
        doc = "instance of EvalFunc" ) )

  #-----------------------------------------------------------------------------
  @property
  def supported( self ):
    return self._supported

  #-----------------------------------------------------------------------------
  def check( self, src ):

    if isinstance( src, EvalFunc ):
      return self.supported['call_py'], src._func

    if isinstance( src, str ):

      if src.startswith(self.TAG_EXPR_PY):
        # python expression
        return self.supported['expr_py'], src[len(self.TAG_EXPR_PY):].lstrip()

      elif src.startswith("$expr:py"):
        # python expression (legacy)
        return self.supported['expr_py'], src[len("$expr:py"):].lstrip()

      elif src.startswith(self.TAG_FUNC_PY):
        # python function
        return self.supported['func_py'], src[len(self.TAG_FUNC_PY):]

      elif src.startswith("$func:py"):
        # python function (legacy)
        return self.supported['func_py'], src[len("$func:py"):]

    return None

  #-----------------------------------------------------------------------------
  def escaped( self, support, src ):
    if support is self.supported['call_py']:
      return EvalFunc( func = src )

    elif support is self.supported['expr_py']:
      # python expression
      return self.TAG_EXPR_PY + src

    elif support is self.supported['func_py']:
      # python function
      return self.TAG_FUNC_PY + src

    raise ValueError(f"`support` must be one of {self.supported.values()}: {support}")

  #-----------------------------------------------------------------------------
  def eval( self,
    schema,
    src,
    locals = None,
    module = None,
    logger = None ):

    if locals is None:
      locals = dict()

    if logger is None:
      logger = log

    supported = self.check( src )

    if not supported:
      raise SchemaEvaluationError(f"No evaluation support: {src}")

    support, src = supported

    # ensure currend directory is preserved after executing script
    cwd = os.path.abspath( os.getcwd() )

    try:

      if support == self.supported['call_py']:
        return src( **locals )

      elif support in [ self.supported['expr_py'], self.supported['func_py'] ]:

        def _print(*objects, sep=' ', end='\n', file=sys.stdout, flush=False):


          if file is sys.stdout:
            msg = sep.join([ str(o) for o in objects])
            logger.info( msg )
          elif file is sys.stderr:
            msg = sep.join([ str(o) for o in objects])
            logger.error( msg )
          else:
            print( *objects, sep=sep, end=end, file=file, flush=flush )

        if module is None:
          module = schema.__module__

        if isinstance( module, str ):
          module = importlib.import_module(module)

        gvars = {
          '__name__' : module.__name__,
          '__file__' : module.__file__ if hasattr(module, '__file__') else None,
          '__builtins__' : {
            **__builtins__,
            'print' : _print } }

        if support == self.supported['expr_py']:
          # python expression

          lvars = dict(locals)

          return eval( src, gvars, lvars )

        elif support == self.supported['func_py']:
          # python function

          lvars = dict(
            __result = None,
            __kwargs = locals )

          # add indentation
          src = "\n".join([ ("  " + l) for l in src.splitlines() ])

          var_keys = ", ".join([ k for k,v in locals.items() ])

          filename = "__src_code_py"
          funcname = "__func_py"

          src = f"\ndef {funcname}({var_keys}):\n{ src }\n\n__result = {funcname}(**__kwargs)"

          try:
            src_code = compile( src, filename, "exec" )

          except BaseException as e:
            raise SchemaEvaluationError(
              f"Error in Python compile",
              hints = SchemaHint.cast( e ) ) from e

          try:
            exec( src_code, gvars, lvars )

          except BaseException as e:
            ehint = extract_exec_stack( e, filename, funcname )

            raise SchemaEvaluationError(
              f"Error in Python exec",
              hints = ehint ) from e

          return lvars['__result']


      assert False, str(support)

    finally:
      os.chdir( cwd )

  #-----------------------------------------------------------------------------
  def lint( self,
    schema,
    src,
    locals = None,
    module = None,
    lint = None,
    logger = None ):

    supported = self.check( src )

    if not supported:
      raise SchemaEvaluationError(f"No evaluation support: {src}")

    support, src = supported

    if support == self.supported['call_py']:
      return list()

    elif support == self.supported['expr_py']:

      return self.lint_python( src, locals, embed_func = False )

    elif support == self.supported['func_py']:

      return self.lint_python( src, locals, embed_func = True )

  #-----------------------------------------------------------------------------
  def lint_python( self, src, locals, embed_func ):
    return lint_python(src, locals, embed_func)



#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
py_provider = PyProvider()

class PyEvaluated( Evaluated, provider = py_provider ):
  pass



import logging
log = logging.getLogger( __name__ )

from copy import copy

from partis.utils import (
  adict,
  adict_struct,
  odict,
  attrs_modify )


from partis.utils.special import (
  required,
  optional )

from partis.schema_meta.valued import (
  ValuedMeta )

from partis.schema_meta.base import (
  SchemaError,
  SchemaHint,
  Loc,
  assert_valid_name,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_special,
  is_optional,
  is_required,
  is_derived,
  is_similar_value_type,
  is_schema_prim,
  is_schema_declared,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_evaluated_class,
  is_evaluated,
  is_valued,
  is_valued_type,
  any_schema )

from partis.schema import (
  EvaluatedContext )

from .valued import (
  Valued,
  get_schema_loc,
  get_init_val,
  get_src_val )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class IntValued( Valued, int ):
  """Container for fully or partially evaluated integer data

  Parameters
  ----------
  val : int
    Source data to be evaluated
  schema : :class:`Schema <partis.schema.struct.SchemaStruct>` | :class:`SchemaPrim <partis.schema.prim.base.SchemaPrim>`
  loc : :class:`Loc <partis.schema_meta.base.Loc>`
    Location information of source data (E.G. file, line/column number)
  """

  #-----------------------------------------------------------------------------
  def __new__( cls,
    val = None,
    schema = None,
    loc = None ):

    val = get_init_val(
      val = val,
      schema = schema,
      default_val = int() )

    self = int.__new__( cls, val )

    return self

  #-----------------------------------------------------------------------------
  def __init__( self,
    val = None,
    schema = None,
    loc = None ):

    int.__init__( self )

    Valued.__init__( self,
      val = val,
      schema = schema,
      loc = loc )

  #-----------------------------------------------------------------------------
  def __str__( self ):

    if is_evaluated( self._src ):
      return str(self._src)

    return str(int(self))

  #-----------------------------------------------------------------------------
  def __repr__( self ):
    return str(self)

  #-----------------------------------------------------------------------------
  @property
  def _encode( self ):
    if self._valued:
      return int(self)

    if is_evaluated( self._src ):
      return self._src._encode

    assert(False)




__all__ = [
  "base",
  "eval",
  "prim",
  "property",
  "schema",
  "struct",
  "valued"  ]

import sys
from partis.utils.module import LazyModule

self = sys.modules[ __name__ ]
self.__class__ = LazyModule

self.define(
  children = __all__ )
